#pragma once

#include <limits>
#include <iostream>
#include <cstring>
#include <cassert>
#include <vector>

namespace Container {


template<int logK, typename keyType>
class KHeapMax {

public:

	// Expose types
	typedef keyType KeyType;

	// Some constants regarding the heap
	static const int NULLINDEX = 0xFFFFFFFF;
	static const int K = 1 << logK;

	// A struct defining a heap element.
	struct HeapElement {
		KeyType key;
		int element;
		HeapElement() : key(0), element(0) {}
		HeapElement(const KeyType k, const int e) : key(k), element(e) {}
	};


public:

	// Constructor of the heap
	KHeapMax(int n) : maxNumberOfElements(n), heap(NULL), position(NULL) {
		position = new int[n];
		heap = new HeapElement[n];
		Initialize();
	}

	// Destroy the heap
	~KHeapMax() {
		if (position != NULL)
			delete[] position;
		if (heap != NULL)
			delete[] heap;
	}


	// Size of the heap
	inline int Size() const {
		return numberOfElements;
	}

	// Heap empty?
	inline bool Empty() const {
		return Size() == 0;
	}

	// extract max element
	inline void ExtractMax(int &element, KeyType &key) {
	  //Assert(!Empty());

		HeapElement &front = heap[0];

		// Assign element and key
		element = front.element;
		key = front.key;

		// Replace heap[0] by last element
		position[element] = NULLINDEX;
		--numberOfElements;
		if (!Empty()) {
			front = heap[numberOfElements];
			position[front.element] = 0;
			SiftDown(0);
		}
	}

	// pop max element.
	inline void PopMax() {
	  //Assert(!Empty());

		HeapElement &front = heap[0];

		// Assign element and key
		int element = front.element;
		KeyType key = front.key;

		// Replace heap[0] by last element
		position[element] = NULLINDEX;
		--numberOfElements;
		if (!Empty()) {
			front = heap[numberOfElements];
			position[front.element] = 0;
			SiftDown(0);
		}
	}


	// delete an arbitrary element from the heap
	inline void Delete(const int element) {
		// If the element is not in the heap, do nothing.
		if (!Contains(element))	return;

		// Get a reference to the heap element.
		int elementPosition = position[element];
		HeapElement &heapElement = heap[elementPosition];
		KeyType previousKey = heapElement.key;

		// Remove the element
		position[element] = NULLINDEX;
		--numberOfElements;
		if (elementPosition < numberOfElements) {
			// Copy the last element from the heap here.
			heapElement = heap[numberOfElements];
			position[heapElement.element] = elementPosition;
			// If the new key shrinked, sift down.
			if (heapElement.key < previousKey)
				SiftDown(elementPosition);
			// Otherwise, sift up.
			else
				SiftUp(elementPosition);
		}
	}


	// update an element of the heap
	inline void Update(const int element, const KeyType key) {
		if (position[element] == NULLINDEX) {
			HeapElement &back = heap[numberOfElements];
			back.key = key;
			back.element = element;
			position[element] = numberOfElements;
			SiftUp(numberOfElements++);
		} else {
			int el_pos = position[element];
			HeapElement &el = heap[el_pos];
			if (key < el.key) {
				el.key = key;
				SiftDown(el_pos);
			} else {
				el.key = key;
				SiftUp(el_pos);
			}
		}
	}


	// Get min key
	inline KeyType MaxKey() const {
		/* Assert(!Empty()); */
		return heap[0].key;
	}

	// Get min element
	inline int MaxElement() const {
		/* Assert(!Empty()); */
		return heap[0].element;
	}

	// Get min element+key
	inline void Max(int &element, KeyType &key) {
		/* Assert(!Empty()); */
		key = heap[0].key;
		element = heap[0].element;
	}

	// Get the key of an arbitrary element
	inline KeyType Key(const int element) const {
		/* Assert(Contains(element)); */
		return heap[position[element]].key;
	}

	// Clear the heap
	inline void Reset() {
		Initialize();
	}

	// Flush the heap
	inline void Clear() {
		for (int i = 0; i < numberOfElements; ++i) {
			position[heap[i].element] = NULLINDEX;
		}
		numberOfElements = 0;
	}

	// Test whether an element is contained in the heap?
	inline bool Contains(const int element) const {
		return position[element] != NULLINDEX;
	}


	// Dumps the heap into a stream.
    void Dump(std::ostream &os) const {
		os << "Heap: ";
		for (int i = 0; i < Size(); ++i) {
            std::cout << "[" << i << ":" << heap[i].key << "] " << std::flush;
		}
        os << std::endl;
		os << "Pos:  ";
		for (int e = 0; e < maxNumberOfElements; ++e) {
			if (position[e] != NULLINDEX) {
                os << "[" << e << ":" << position[e] << "] " << std::flush;
			}
		}
        os << std::endl;
	}

protected:

	// Reset the heap
	inline void Initialize() {
		numberOfElements = 0;
		memset(position, 0xFF, sizeof(int) * maxNumberOfElements);
		for (int i = 0; i < maxNumberOfElements; ++i) {
			assert(position[i] == NULLINDEX);
		}
	}

	// Sift up an element
	inline void SiftUp(int i) {
		assert(i < numberOfElements);
		int cur_i = i;
		while (cur_i > 0) {
			int parent_i = (cur_i-1) >> logK;
			if (heap[parent_i].key < heap[cur_i].key)
				Swap(cur_i, parent_i);
			else
				break;
			cur_i = parent_i;
		}
	}

	// Sift down an element
	inline void SiftDown(int i) {
		assert(i < numberOfElements);

		while (true) {
			int min_ind = i;
			KeyType min_key = heap[i].key;

			int child_ind_l = (i << logK) + 1;
			int child_ind_u = std::min(child_ind_l + K, numberOfElements);

			for (int j = child_ind_l; j < child_ind_u; ++j) {
				if (heap[j].key > min_key) {
					min_ind = j;
					min_key = heap[j].key;
				}
			}

			// Exchange?
			if (min_ind != i) {
				Swap(i, min_ind);
				i = min_ind;
			} else {
				break;
			}
		}
	}

	// Swap two elements in the heap
	inline void Swap(const int i, const int j) {
		HeapElement &el_i = heap[i];
		HeapElement &el_j = heap[j];

		// Exchange positions
		position[el_i.element] = j;
		position[el_j.element] = i;

		// Exchange elements
		HeapElement temp = el_i;
		el_i = el_j;
		el_j = temp;
	}



private:
	// Number of elements in the heap
	int numberOfElements;

	// Number of maximal elements
	int maxNumberOfElements;

	// Array of length heap_elements.
	HeapElement *heap;

	// An array of positions for all elements.
	int *position;
};

}

