#include "instance.hpp"
#include "solution.hpp"
#include "solution_algo.hpp"
#include "geo.hpp"
#include "io.hpp"
#include "collision.hpp"
#include "greedy.h"
#include "tools.hpp"

#include <sys/time.h>

#include <cstring>
#include <iostream>
#include <map>
#include <fstream>
#include <boost/program_options.hpp>

using namespace std;


// Perform schematization by path files from a directory
bool load_file(const string &filename, ProgramState& state) {
  
  state.reset();
  if(!load_cl(filename,state)) {
    cout << "Error loading file" <<endl;
    return false;
  }

  return true;
}

int nmb_active_intervals(const Solution& sol, const Instance& inst){
  int nmb_intervals = 0;
  for (unsigned int i = 0; i < inst.count_labels(); i++){
    AngleSet active = sol.get_active(i);
    //    nmb_intervals += (active.events.size()/2);
    if (active == AngleSet::empty)
      continue;

    if (active.events.size() == 2){
      nmb_intervals++;
      continue;
    }
    
    nmb_intervals += (active.events.size()/2);
        
    if ((active.events[0] == 0.0) && (active.events[active.events.size() - 1] == 2*M_PI))
      nmb_intervals--;
  }

  return nmb_intervals;
}


struct Experiment_configuration {
    
    string filename = {""};
    int max_nmb_intervals = 1;
    
    bool use_hard_conflicts = false;
    bool use_optimal = false;
    bool use_optimalp = false;
    bool use_optimal2 = false;
    bool use_optimal2p = false;
    bool use_min_length = false;
    bool use_min_length_ilp = false;
    bool use_greedy_max = false;
    bool use_greedy_br = false;
    bool use_greedy_lc = false;
    bool use_quarter = false;
    bool use_quarter_greedy = false;
    bool use_no_flickering = false;
    
    ProgramState state;
    map<string, string> stat_files;
    
    vector<string> algonames;
};

struct Experiment_results {
    map <string, double> times;
    map <string, double> objval;
    map <string, double> avg_length;
    map <string, double> avg_count;
    map <string, double> gap;
};


inline void get_stats(string algoname, double time, ProgramState& state, Experiment_results& results) {

  results.times [algoname] = time;
  results.objval[algoname] = state.solution->active_time();
  int nmb_intervals = nmb_active_intervals(*(state).solution, *(state).instance);
  results.avg_length[algoname] = state.solution->active_time() / (double)nmb_intervals;
  results.avg_count[algoname] = (double) nmb_intervals / state.instance->count_labels() ;
  results.gap[algoname] = state.solution->gap;
}



void run_algorithm(const string& algoname, const string& filename, Experiment_configuration& config, Experiment_results& results) {
    //std::cout << algoname << std::endl;
    Timer t;
    t.restart();
    
    AlgorithmInfo algo;
    algo.max_nmb_intervals = config.max_nmb_intervals;
    algo.use_hard_conflicts = config.use_hard_conflicts;
    algo.minimum_length = 0.0;
    algo.time_limit = 3600;
    
    std::cout << algoname << std::endl;

    if (algoname == "greedy max       ")
        algo.name = AlgorithmInfo::GreedyMax;
    else if (algoname == "greedy low cost  ")
        algo.name = AlgorithmInfo::GreedyLowCost;
    else if (algoname == "greedy best ratio")
        algo.name = AlgorithmInfo::GreedyBestRatio;

    if ( (algoname == "greedy max       ") || (algoname == "greedy low cost  ") || (algoname == "greedy best ratio") )
        config.state.solution = new Solution(calc_greedy(*(config.state).instance, algo));
    else if (algoname == "optimal          ") {
        algo.name = AlgorithmInfo::MIP;
        config.state.solution = new Solution(calc_mip_solution_fast(algo, *(config.state).instance, false));
    }
    else if (algoname == "optimal presolve ") {
        algo.name = AlgorithmInfo::MIP;
        config.state.solution = new Solution(calc_mip_solution_fast(algo, *(config.state).instance, true));
    }
    else if (algoname == "optimal2         ") {
        algo.name = AlgorithmInfo::MIP2;
        config.state.solution = new Solution(calc_mip_solution_fast(algo, *(config.state).instance, false));
    }
    else if (algoname == "optimal2p        ") {
        algo.name = AlgorithmInfo::MIP2;
        config.state.solution = new Solution(calc_mip_solution_fast(algo, *(config.state).instance, true));
    }
    else if (algoname == "no flickering    ") {
        config.state.solution = new Solution(calc_mip_solution_no_flickering(*(config.state).instance, config.use_hard_conflicts));
    }
    else if (algoname == "qapx             ") {
        config.state.solution = new Solution(calc_app(*(config.state).instance, config.max_nmb_intervals, config.use_hard_conflicts));
    }
    else if ( (algoname == "qapx+gm          ") || (algoname == "qapx+glc         ") || (algoname == "qapx+gbr         ") ) {
        vector <Solution> apx_solutions;
        //first calculate optimal solution for each of the four subinstances
        t.restart();
        algo.name = AlgorithmInfo::MIP;

        calc_appr(*(config.state).instance, algo, 4, &apx_solutions);

        (config.state).solution = new Solution(apx_solutions[0]);
        assert(apx_solutions.size() == 4);

        for (unsigned int i = 0; i < apx_solutions.size(); i++){
            Solution sol = apx_solutions[i];

            AlgorithmInfo algo_greedy(algo);
            if (algoname == "qapx+gm          ")
                algo_greedy.name = AlgorithmInfo::GreedyMax;
            else if (algoname == "qapx+glc         ")
                algo_greedy.name = AlgorithmInfo::GreedyLowCost;
            else if (algoname == "qapx+gbr         ")
                algo_greedy.name = AlgorithmInfo::GreedyBestRatio;

            Greedy g(algo_greedy, *(config.state).instance, sol, true);
            
            if (sol.active_time() > config.state.solution->active_time())
                (config.state).solution = new Solution(sol);
        }
    }
    else
        assert(false);
    
    double time = t.elapsed();

    get_stats(algoname, time, config.state, results);
    
    int nmb_intervals = nmb_active_intervals(*(config.state).solution, *(config.state).instance);
    ofstream output(config.stat_files[algoname].c_str(), std::fstream::app);
    output <<  filename << ";" << config.state.instance->count_labels() << ";" << config.max_nmb_intervals << ";" << (config.use_hard_conflicts ? "1" : "0") << ";";
    output << config.state.solution->active_time() << ";" << time << ";" <<  config.state.solution->gap << ";";
    output <<  config.state.solution->active_time() / (double)nmb_intervals << ";" << (double) nmb_intervals / config.state.instance->count_labels()  << "\n";
    output.close();
    string sol_outfilename = filename + "." + algoname + ".sol";
    save(sol_outfilename, config.state);
}

int main(int argc, char *argv[]){
    
  Experiment_configuration config;

  string unique_id("");

  namespace po = boost::program_options;
  po::options_description desc("Options");
  desc.add_options()
    ("help,h", "displays this screen")
    ("file,f", po::value<string>(&config.filename), "filename")
    ("intervals,i", po::value<int>(&config.max_nmb_intervals), "maximum number of intervals [default 1]")
    ("optimal,o", po::value( &config.use_optimal )->zero_tokens(), "use optimal algorithm")
    ("optimalp,p", po::value( &config.use_optimalp )->zero_tokens(), "use optimal algorithm with initial start configuration obtained by greedy max")
    ("optimal2,x", po::value( &config.use_optimal2 )->zero_tokens(), "use second optimal algorithm")
    ("optimal2p,y", po::value( &config.use_optimal2p )->zero_tokens(), "use second optimal algorithm with initial start configuration obtained by greedy max")
    ("noflickering,n", po::value( &config.use_no_flickering )->zero_tokens(), "no flickering")
    ("greedy,g", "use all greedy algorithms")
    ("greedymax,m", po::value( &config.use_greedy_max )->zero_tokens(), "use greedy max")
    ("greedylc,l", po::value( &config.use_greedy_lc )->zero_tokens(), "use greedy low cost")
    ("greedybr,b", po::value( &config.use_greedy_br )->zero_tokens(), "use greedy best ratio")
    ("qapx,q", po::value( &config.use_quarter )->zero_tokens(), "use quarter approximation")
    ("qapxe,e", po::value( &config.use_quarter_greedy )->zero_tokens(), "use quarter approximation and enrich solution with all three greedy algorithms.")
    ("hc,c", po::value( &config.use_hard_conflicts )->zero_tokens(), "use hard conflicts")
    ("minlength,z", po::value( &config.use_min_length )->zero_tokens(), "use max min length optimal")
    ("minlengthilp,r", po::value( &config.use_min_length_ilp )->zero_tokens(), "use max min length optimal + maximize with ILP")
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);

  if (vm.count("help")) {
    std::cout << desc << "\n";
    return false;
  }

  if (config.filename == "") {
    std::cout << "no filename (-f) specified!" << std::endl;
    return false;
  }
    
  if (config.max_nmb_intervals <= 0.0)
    config.max_nmb_intervals = -1.0;
  
    if (vm.count("greedy")) {
        config.use_greedy_br = config.use_greedy_lc = config.use_greedy_max = true;
    }



  cout << "experiment for " << config.filename << endl;
  cout << "algorithms: " 
       << (config.use_optimal ? "opt, " : "" )
       << (config.use_greedy_max ? "greedy max, " : "")
       << (config.use_greedy_lc ? "greedy low cost, " : "")
       << (config.use_greedy_br ? "greedy best ratio, " : "")
       << (config.use_quarter ? "quarter apx": "" )
       << (config.use_quarter_greedy ? "quarter apx + all greedy combinations": "")
       << (config.use_min_length ? "min length" : "")
       << endl;
  cout << "maximum number of intervals: " << config.max_nmb_intervals << endl;
  cout << "hard conflicts " << ( (config.use_hard_conflicts ? " true " : " false ") ) << endl;



  //GREEDY
  if (config.use_greedy_max)
    config.algonames.push_back("greedy max       ");

  if (config.use_greedy_lc)
    config.algonames.push_back("greedy low cost  ");

  if (config.use_greedy_br) 
    config.algonames.push_back("greedy best ratio");

  //OPTIMAL
  if (config.use_optimal)
    config.algonames.push_back("optimal          ");
  if (config.use_optimalp)
    config.algonames.push_back("optimal presolve ");

  if (config.use_optimal2 && config.max_nmb_intervals == 1)
    config.algonames.push_back("optimal2         ");
  if (config.use_optimal2p && config.max_nmb_intervals == 1)
    config.algonames.push_back("optimal2p        ");

  if (config.use_min_length && config.max_nmb_intervals == 1)
    config.algonames.push_back("min length       ");
  if (config.use_min_length_ilp && config.max_nmb_intervals == 1)
    config.algonames.push_back("min length a ilp ");

  if (config.use_no_flickering)
    config.algonames.push_back("no flickering    ");

  //Quarter Approximation
  if (config.use_quarter)
    config.algonames.push_back("qapx             ");

  //quarter apx + greedy
  if (config.use_quarter_greedy){
      config.algonames.push_back("qapx+gm          ");
      config.algonames.push_back("qapx+glc         ");
      config.algonames.push_back("qapx+gbr         ");
  }

  string out_filename = config.filename + "_stats.csv";

  cout << "Dumping csv-file to " << out_filename << "... " << flush;

  // Try to open the file
  ofstream f(out_filename.c_str());
  if (!f.is_open()) {
    cout << "failed." << endl;
    return 0;
  }
  
  for (const string& algorithm_name: config.algonames) {
    const string out_filename_csv = config.filename + "_" + algorithm_name + "_hc_" + (config.use_hard_conflicts ? "1" : "0") + "_i_" + std::to_string(config.max_nmb_intervals) + "_stats.csv";

    ofstream csv_file(out_filename_csv.c_str());
    csv_file << "filename;nmb_labels;max_nmb_intervals;hc;objval;time;gap;~interval-length;~interval-count;\n";
    csv_file.close();

    config.stat_files[algorithm_name] = out_filename_csv;
  }

 ifstream file_list(config.filename);
 if (file_list.is_open()) {
   string line;
     
   while (getline (file_list,line)){
      cout << line << endl; 
      if (line == "\n")
          continue;

      bool bFileLoaded = load_file(line, config.state);
       
      if (!bFileLoaded) {
          std::cout << "could not open file: " << line << std::endl;
          continue;
      }
      //std::cout << "loaded file" << std::endl;

      Experiment_results results;
      for (string & algorithm_name: config.algonames) {
          run_algorithm(algorithm_name, line, config, results);
      }
       //std::cout << "finished running all selected algorithms" << std::endl;

   }
    file_list.close();
  }


  f.close();

  cout << "done." << endl;
  
  return 0;
}
		  
