#ifndef GEO_HPP
#define GEO_HPP

#include <cmath>

#define EULER 2.718281828459045
#ifndef M_PI
#define M_PI 3.141592653589793
#endif

//mercator projection
inline double y2lat(double a){
  return 180.0/M_PI*(2.0*atan(pow(EULER, (a*M_PI/180.0))-M_PI/2.0));
}		     
inline double lat2y(double a){
  return 6378137.0*log(tan(M_PI/4.0+a*(M_PI/180.0)/2.0));
}
inline double lon2x(double a){
  return a * (M_PI/180.0) * 6378137.0;
}
inline double x2lon(double a){
  return a / (M_PI/180.0) / 6378137.0;
}

inline double latlongdist(double lat1, double lon1, double lat2, double lon2){
  double R = 6371; // radius of earth (km)
  double dLat = (lat2-lat1);
  double dLon = (lon2-lon1);

  double a =  sin(dLat/2) * sin(dLat/2) + 
    sin(dLon/2) * sin(dLon/2) * cos(lat1) * cos(lat2); 
  double c = 2 * atan2(sqrt(a), sqrt(1-a)); 
  double d = R * c;
  return d;
}


struct Vector{
	Vector(){}
	Vector(double x, double y):
		x(x), y(y){}
	double x, y;

	Vector operator-()const{
		return Vector(-x, -y);
	}

	//comparison for ordering in a map
	bool operator<(Vector other)const{
		return this->x < other.x || (this->x == other.x && this->y < other.y);
	}
  
  bool operator==(const Vector& other)const {
    return this->x == other.x && this->y == other.y;
  }
};

inline Vector operator-(Vector a, Vector b){
	return Vector(a.x-b.x, a.y-b.y);
}

inline Vector operator+(Vector a, Vector b){
	return Vector(a.x+b.x, a.y+b.y);
}

inline double norm(Vector v){
	return std::sqrt(v.x*v.x + v.y*v.y);
}

struct Matrix{
	Matrix(){}

	Matrix(double s): a(s), b(0), c(0), d(s){}

	Matrix(double a, double b, double c, double d):
		a(a), b(b), c(c), d(d){}

	double a, b,c,d;


};

inline Vector operator*(Matrix m, Vector v) {
	return Vector(
		m.a*v.x + m.b*v.y, 
		m.c*v.x + m.d*v.y
	);
}

inline Matrix operator*(Matrix m, Matrix n) {
	return Matrix(
		m.a*n.a + m.b*n.c , m.a*n.b + m.b*n.d,
		m.c*n.a + m.d*n.c , m.c*n.b + m.d*n.d
	);
}

struct AffTrans{
	AffTrans(){}

	AffTrans(Matrix m, Vector t = Vector(0.0, 0.0)):
		m(m), t(t){}

	AffTrans(double a, double b, double c, double d, double e, double f):
		m(a, b, d, e), t(c, f){}

	Matrix m;
	Vector t;
};

inline Vector operator*(AffTrans a, Vector v){
	v = a.m*v;
	return Vector(v.x + a.t.x, v.y + a.t.y);
}

inline AffTrans operator*(AffTrans a, AffTrans b){
	return AffTrans(a.m * b.m, a.m*b.t + a.t);
}

namespace geo_util{
	inline Matrix id(){
		return Matrix(
			1, 0, 
			0, 1
		);
	}

	inline Matrix rot(double a){
		double c = std::cos(a), s = std::sin(a);
		return Matrix(
			c, -s, 
			s, c
		);
	}

	inline double det(Matrix m){
		return m.a*m.d - m.b*m.c;
	}

	inline Matrix inv(Matrix m){
		return (1.0/det(m))*Matrix(m.d, -m.b, -m.c, m.a);
	}

	inline AffTrans inv(AffTrans a){
		Matrix m = inv(a.m);
		return AffTrans(m, -(m*a.t));
	}

	inline AffTrans translate(Vector t){
		return AffTrans(id(), t);
	}

	//!< m must be an isometrie
	inline double rot_angle(Matrix m){
		return std::atan2(m.c, m.a);
	}
}

#endif

