#ifdef GREEDY_H
#else
#define GREEDY_H

#include <limits>

#include "angle_set.hpp"
#include "instance.hpp"
#include "solution.hpp"
#include "collision.hpp"
#include "KHeapMax.h"
#include "algorithms.hpp"

#ifndef DEBUG_LINE
#define DEBUG_LINE std::cout << __PRETTY_FUNCTION__ << ", " << __LINE__ << std::endl;
#endif

struct cost_information {
    cost_information(const Instance* const inst, const std::vector<AngleSet>* const free_angles, std::vector<int>* const nmb_intervals_left) :
        inst(inst), free_angles(free_angles), nmb_intervals_left(nmb_intervals_left)
    {
    }

    const Instance* const inst;
    const std::vector<AngleSet>* const free_angles;
    const std::vector<int>* const nmb_intervals_left;

    int label_index;
    const AngleInterval* largest_interval;

    double get_weight() {
        return inst->get_label(label_index).weight;
    }

};

class Greedy{
 public:
    Greedy(const AlgorithmInfo& algorithm, const Instance& instance, Solution& solution, bool use_precomputed_solution = false) {
        algo = &algorithm;
        inst = &instance;
        sol = &solution;
        
        int nmb_labels = (*inst).count_labels();

        free_angles.resize(nmb_labels, AngleSet::all);
        nmb_intervals_left.resize(nmb_labels, (*algo).max_nmb_intervals);

        if ((*algo).use_hard_conflicts)  {

            for (int i = 0; i < nmb_labels; ++i) {
                free_angles[i] = ~(*inst).get_hard_conflicts(i);
            }
        }



        if (use_precomputed_solution ) {
            //\note this implicitly assumes that if a precomputed solution is used, a label may not use any more intervals
            for (int i = 0; i < nmb_labels; ++i) {
              if ((*sol).get_active(i) != AngleSet::empty) {
                        nmb_intervals_left[i] = 0;
                        //free_angles[i] = free_angles[i] & ~((*sol).get_active(i));
                        free_angles[i] = AngleSet::empty;
              }
            }

            // update the free angles for the remaining labels
            for (const conflict_information& c_info: (*inst).get_all_conflicts()) {
                int label_id1 = c_info.label_id1;
                int label_id2 = c_info.label_id2;

                const AngleSet& active_id1 = (*sol).get_active(label_id1);
                const AngleSet& active_id2 = (*sol).get_active(label_id2);

                if (active_id1 != AngleSet::empty)
                    free_angles[label_id2] = free_angles[label_id2] & ( ~(c_info.conflict_angles & active_id1) );

                if (active_id2 != AngleSet::empty)
                    free_angles[label_id1] = free_angles[label_id1] & ( ~(c_info.conflict_angles & active_id2) );
            }
        }

        switch ((*algo).name) {
            case AlgorithmInfo::GreedyMax:
                greedy(Greedy::computeLength);
                break;
            case AlgorithmInfo::GreedyLowCost:
                greedy(Greedy::computeCost);
                break;
            case AlgorithmInfo::GreedyBestRatio:
                greedy(Greedy::computeRatio);
                break;
            case AlgorithmInfo::GreedyShuffle:
                shuffle();
                break;
            default:
                std::cout << (*algo).name << std::endl;
                assert(false);
                break;
        }
        
    }
    
 private:
    
    void shuffle(){
        
        int nmb_labels = (*inst).count_labels();
        double minimum_length = (*algo).minimum_length * 2 * M_PI;
        
        //fill vector from 0...nmb_labels - 1
        std::vector<int> label_indices(nmb_labels);
        int i = 0;
        std::generate (label_indices.begin(), label_indices.end(), [&]{return i++;});
        
        //shuffle vector
        std::random_device rd;
        std::mt19937 g(rd());
        std::shuffle (label_indices.begin(), label_indices.end(), g);
        
        //run alg
        std::vector <AngleSet> temp_sol(nmb_labels, AngleSet::empty);

        bool made_changes = true;
        while (made_changes) {
            made_changes = false;
            
            for (const int& index: label_indices) {

                if (nmb_intervals_left[index] == 0)
                    continue;
                
                AngleInterval interval = find_largest_sub_interval(free_angles[index]);

                if (length(interval) > minimum_length) {
                    AngleSet active_range(interval);
                    temp_sol[index] = temp_sol[index] | active_range;
                    nmb_intervals_left[index]--;
                    
                    //update
                    free_angles[index] = free_angles[index] & (~active_range);
                    
                    //update collision partners
                    std::vector<conflict_information> conflicts = (*inst).get_all_conflicts_for_label(index);
                    for (const conflict_information& cinfo: conflicts) {
                        int other_label = cinfo.get_other_label_id(index);

                        if (nmb_intervals_left[other_label] == 0)
                            continue;

                        free_angles[other_label] = (~(active_range & cinfo.conflict_angles) & free_angles[other_label]);
                    }
                    made_changes = true;
                }
                else
                    nmb_intervals_left[index] = 0;
                
            }
        }

        //copy solution
        for (int i = 0; i < nmb_labels; ++i){
            if ((*sol).get_active(i) == AngleSet::empty)
                (*sol).set_active(i, temp_sol[i]);
        }
    }
    
    template <typename T>
    void update_heap(T ratingFunc, Container::KHeapMax<1, double>& kheap, int label_index) {
        AngleInterval largest_interval = find_largest_sub_interval(free_angles[label_index]);

        if (length(largest_interval) > (*algo).minimum_length * 2 * M_PI) {
            cost_information cost_info(inst, &free_angles, &nmb_intervals_left);
            cost_info.label_index = label_index;
            cost_info.largest_interval = &largest_interval;

            kheap.Update(label_index, ratingFunc(cost_info));
        }
    }

    template <typename T>
    void greedy(T ratingFunc){

        int nmb_labels = (*inst).count_labels();
        
        //fill heap
        Container::KHeapMax<1, double> kheap(nmb_labels);
        for (int i = 0; i < nmb_labels; ++i){
            if (nmb_intervals_left[i] == 0)
                continue;

            update_heap(ratingFunc, kheap, i);
        }

        std::vector <AngleSet> temp_sol(nmb_labels, AngleSet::empty);

        int index = 0;
        double heap_value = 0.0;
        while (!kheap.Empty()){
            //choose label
            kheap.ExtractMax(index, heap_value);
            assert (nmb_intervals_left[index] != 0);

            AngleInterval interval = find_largest_sub_interval(free_angles[index]);
            
            //set solution
            AngleSet active_range(interval);
            temp_sol[index] = temp_sol[index] | active_range;
            --nmb_intervals_left[index];
            
            //update data for label (if necessary)
            if (nmb_intervals_left[index] != 0){
                free_angles[index] = free_angles[index] & (~active_range);
                update_heap(ratingFunc, kheap, index);
            }
            
            //update collision partners
            std::vector<conflict_information> conflicts = (*inst).get_all_conflicts_for_label(index);
            std::vector<int> label_ids;
            for (const conflict_information& cinfo: conflicts) {
                int other_label = cinfo.get_other_label_id(index);

                if (nmb_intervals_left[other_label] == 0)
                    continue;

                free_angles[other_label] = (~(active_range & cinfo.conflict_angles) & free_angles[other_label]);
                label_ids.push_back(other_label);
            }

            //note: this second loop cannot(!) be joined with the loop above, because *all* updates to free_angles must be done first
            for (int label_index: label_ids)
                update_heap(ratingFunc, kheap, label_index);
        }
        
        //copy solution
        for (int i = 0; i < nmb_labels; ++i){
            if ((*sol).get_active(i) == AngleSet::empty)
                (*sol).set_active(i, temp_sol[i]);
        }
    }
    
    static inline double determine_cost(cost_information& cost_info) {
        double cost = 0;
        AngleSet dummy = AngleSet(*cost_info.largest_interval);

        //iterate over all conflicts
        std::vector<conflict_information> conflicts = cost_info.inst->get_all_conflicts_for_label(cost_info.label_index);
        for (const conflict_information& cinfo: conflicts) {
            int other_label = cinfo.get_other_label_id(cost_info.label_index);

            if (cost_info.nmb_intervals_left->at(other_label) == 0)
                continue;

          AngleInterval largest_interval_old = find_largest_sub_interval(cost_info.free_angles->at(other_label));
          AngleInterval largest_interval_new = find_largest_sub_interval(cost_info.free_angles->at(other_label) & dummy & cinfo.conflict_angles);
          double weight = cost_info.get_weight();
          cost += (length(largest_interval_old) - length(largest_interval_new) ) * weight;
        }
        return cost;
      }


    static inline double computeLength(cost_information& cost_info){
        double weight = cost_info.get_weight();
        return length(*cost_info.largest_interval) * weight;
    }

    static inline double computeCost(cost_information& cost_info){
        double cost = determine_cost(cost_info);
        if (cost == 0)
            return std::numeric_limits<double>::max();

        return 1.0 / cost;
    }

   static inline double computeRatio(cost_information& cost_info){
       double cost = determine_cost(cost_info);
       double weight = cost_info.get_weight();

       double ratio = std::numeric_limits<double>::max();
       if (cost > 0)
           ratio = (length(*cost_info.largest_interval) / cost) * weight;

       return ratio;
   }

private:
   std::vector<AngleSet> free_angles;
   std::vector <int> nmb_intervals_left;

   const Instance* inst = nullptr;
   const AlgorithmInfo* algo = nullptr;
   Solution* sol = nullptr;
};

#endif
