#ifndef INSTANCE_HPP
#define INSTANCE_HPP

#include "geo.hpp"
#include "label.h"
#include "angle_set.hpp"
#include "collision.hpp"

#include <string>
#include <cstdio>
#include <vector>
#include <cmath>
#include <iomanip>


struct conflict_information {
  unsigned int label_id1;
  unsigned int label_id2;

conflict_information(unsigned int id1, unsigned int id2, const AngleSet& a) :
  label_id1(id1), label_id2(id2), conflict_angles(a) {
  }

  unsigned int get_other_label_id(unsigned int label_id) const {
    assert(label_id == label_id1 || label_id == label_id2);
    return (label_id == label_id1) ? label_id2 : label_id1;
  }

  AngleSet conflict_angles = AngleSet::empty;
};

class Instance{
public:
    // label handling

	unsigned count_labels() const { return labels.size(); }
	const Label&get_label(unsigned i)const{return labels[i];}
	Label&get_label(unsigned i){return labels[i];}

    unsigned add_label(const Label&l) {
        labels.push_back(l);
        labels[labels.size()-1].update();
        return labels.size()-1;
    }

    void remove_label(unsigned i) {
        labels.erase(labels.begin()+i);
    }

    //conflict handling
    void compute_conflicts() {

        for (std::vector<int>& v: conflict_indices)
            v.clear();
        conflict_indices.resize(count_labels());

        soft_conflict_info.clear();
        hard_conflict_info.clear();
        hard_conflict_info.resize(count_labels(), AngleSet::empty);



        for (AngleSet& a: hard_conflict_info)
            a = AngleSet::empty;

        //... O(n^2) .. not great!
        for(unsigned int i = 0; i < count_labels(); ++i) {

            for(unsigned j = i + 1; j < count_labels(); ++j) {

                /* compute soft conflicts */
                AngleSet a = calc_collision(get_label(i), get_label(j));

                if (a != AngleSet::empty){
                   soft_conflict_info.emplace_back(i, j, a);

                  conflict_indices[i].push_back(soft_conflict_info.size() - 1);
                  conflict_indices[j].push_back(soft_conflict_info.size() - 1);
                }

                /* compute hard conflicts */
                AngleSet hc_ij = calc_hard_conflicts(get_label(i), get_label(j));
                AngleSet hc_ji = calc_hard_conflicts(get_label(j), get_label(i));

                if (hc_ij != AngleSet::empty)
                    hard_conflict_info[i] = hard_conflict_info[i] | hc_ij;

                if (hc_ji != AngleSet::empty)
                    hard_conflict_info[j] = hard_conflict_info[j] |  hc_ji;

            }
        }
    }

    const AngleSet& get_conflict(unsigned int i, unsigned int j) const {
        assert(conflict_indices.size() == count_labels());

        for (int index: conflict_indices[i]) {
            if (soft_conflict_info[index].get_other_label_id(i) == j) {
                return soft_conflict_info[index].conflict_angles;
            }
        }
        return AngleSet::empty;
    }

    const AngleSet& get_hard_conflicts(unsigned int i) const {
        assert(hard_conflict_info.size() == count_labels()); //should have done preprocessing
        return hard_conflict_info[i];
    }

    const std::vector<conflict_information>& get_all_conflicts() const {
        return soft_conflict_info;
    }

    std::vector<conflict_information> get_all_conflicts_for_label(unsigned i) const {
        std::vector<conflict_information> result;
        for (int conflict_index: conflict_indices[i])
            result.push_back(soft_conflict_info[conflict_index]);

        return result;
    }

private:
	std::vector<Label>labels;

    std::vector<std::vector<int>> conflict_indices;
    std::vector<conflict_information> soft_conflict_info;
    std::vector<AngleSet> hard_conflict_info;

};

Instance*load_instance(const char*);
bool load_instance(std::FILE*f, Instance&);
void save_instance(const char*, const Instance&);
void save_instance(std::FILE*f, const Instance&);
Label random_label();
Instance random_instance(unsigned label_count);

#endif

