#include "instance.hpp"
#include "solution.hpp"
#include "io.hpp"

#include <iostream>

#include <cstdio>
#include <string>
#include <cctype>

using namespace std;

const char file_tag [] = "#labelingRotatingMaps-File-v1.1\n";
const char background_tag [] = "#Background:";
const char font_tag [] = "#Font:";
const char instance_tag [] = "#        x         y     right       top      left    bottom    text";

class IllegalFileFormatError{};

void skip_string(FILE*f, const char*str){
  while(*str){
    int c = fgetc(f);
    
    if(c == EOF || c != *str){
      throw IllegalFileFormatError();
    }
    ++str;
  }
}

void soft_skip_string(FILE*f, const char*str){
	while(*str){
		while(*str && isspace(*str))
			++str;
		int c = fgetc(f);
		while(isspace(c))
			c = fgetc(f);
		if(c == EOF || c != *str){
		  cout << __PRETTY_FUNCTION__ << "; " << __LINE__ << endl;
		  throw IllegalFileFormatError();
		}
		++str;
	}
}


void write_string(FILE*f, const std::string&str){
	fputc('"', f);
	for(int i=0, s=str.size(); i<s; ++i){
		if(str[i] == '\\'){
			fputc('\\', f);
			fputc('\\', f);
		}else if(str[i] == '\"'){
			fputc('\\', f);
			fputc('\"', f);
		}else
			fputc(str[i], f);
	}
	fputc('"', f);
}


string read_string(FILE*f){
	int c = fgetc(f);
	while(isspace(c))
		c = fgetc(f);
	if(c != '\"'){
	  cout << __PRETTY_FUNCTION__ << "; " << __LINE__ << endl;
	  cout << c << endl;
	  throw IllegalFileFormatError();
	}
	string str;
	bool escaped = false;
	for(;;){
		c = fgetc(f);
		if(c == EOF){
		  cout << __PRETTY_FUNCTION__ << "; " << __LINE__ << endl;
		  cout << "EOF!" << endl;
		  throw IllegalFileFormatError();
		}
		else if(c == '\\')
			escaped = true;
		else if(c == '"' && !escaped)
			return str;
		else{
			str.push_back(c);
			escaped = false;
		}
	}
}


void write_angle_set(FILE*f, const AngleSet&angle_set){
	for(int i=0, s=angle_set.events.size(); i<s; ++i){
		fprintf(f, " %8.3lf", angle_set.events[i]*180/M_PI);
	}
}

AngleSet read_angle_set(FILE*f){
	AngleSet angle_set;

	char line[1024*2];
	fgets(line, sizeof(line)-1, f);


	char*float_begin = 0, *pos = line;
	for(;;){
		bool at_end = (*pos == 0);
		if(isspace(*pos) || *pos == 0){
			if(float_begin){
				*pos = 0;
				angle_set.events.push_back(atof(float_begin)*M_PI/180);
				float_begin = 0;
			}
		}else if(!float_begin){
			float_begin = pos;
		}
		
		if(at_end)
			break;
		else
			++pos;
	}

	if(!angle_set.events.empty()){
		if(angle_set.events.front() < 0.0)
			angle_set.events.front() = 0.0;
		
		if(angle_set.events.back() > 2*M_PI)
			angle_set.events.back() = 2*M_PI;
		
        for(unsigned int i=1; i<angle_set.events.size(); ++i){
			if(angle_set.events[i-1] >= angle_set.events[i])
				throw IllegalFileFormatError();
		}
	}
	return angle_set;
}

void save(const std::string&file_name, const ProgramState&state){
	FILE*f = fopen(file_name.c_str(), "w");
	fwrite(file_tag, sizeof(file_tag)-1, 1, f);
	fwrite(background_tag, sizeof(background_tag)-1, 1, f);
	write_string(f, state.background_filename);
	fputc('\n', f);

	fwrite(font_tag, sizeof(font_tag)-1, 1, f);
	write_string(f, state.font_desc);
	fputc('\n', f);
	
	if(state.instance){
		fwrite(instance_tag, sizeof(instance_tag)-1, 1, f);
		if(state.solution)
			fprintf(f, "   active\n");
		else
			fprintf(f, "\n");
		
		for(int i=0, s=state.instance->count_labels(); i<s; ++i){
			const Label&l = state.instance->get_label(i);
			fprintf(f, "%10.3lf%10.3lf%10.3lf%10.3lf%10.3lf%10.3lf", l.pos.x, l.pos.y, 
				l.right, l.top, l.left, l.bottom);
			for(int j=0; j<4; ++j)
				fputc(' ', f);
			write_string(f, l.text.c_str());
			if(state.solution)
				write_angle_set(f, state.solution->get_active(i));
			fputc('\n', f);
		}

	}
	fclose(f);
}



bool load(const std::string&file_name, ProgramState&old_state){
	FILE*f = fopen(file_name.c_str(), "r");

	if(!f){
		return false;
	}
	try{
		char buf_[20];
		ProgramState state;


		skip_string(f, file_tag);

		skip_string(f, background_tag);
		state.background_filename = read_string(f);
		fgets(buf_, sizeof(buf_)-1, f);

		skip_string(f, font_tag);
		state.font_desc = read_string(f);
		fgets(buf_, sizeof(buf_)-1, f);

		state.instance = new Instance;
		
		skip_string(f, instance_tag);
		fgets(buf_, sizeof(buf_)-1, f);

		string buf = buf_;
		if(buf.substr(buf.find_first_not_of(" "), 6) == "active")
			state.solution = new Solution;
		else
			state.solution = 0;

		
		for(;;){
			Label l;
			if(fscanf(f, "%lf%lf%lf%lf%lf%lf", &l.pos.x, &l.pos.y, &l.right, &l.top, &l.left, &l.bottom) != 6){
				break;
			}
			l.text= read_string(f);
			// std::cout << l.text << ", " << std::flush;
			state.instance->add_label(l);
			if(state.solution)
				state.solution->active.push_back(read_angle_set(f));
		}

	
		int c = 0;
		while(c != EOF){
			c = fgetc(f);
			if(!isspace(c) && c != EOF)
				return false;
		}			


		if (old_state.instance != nullptr)
		  delete old_state.instance;

		if (old_state.solution != nullptr)
		  delete old_state.solution;		

		old_state = state;
	}catch(IllegalFileFormatError){
		fclose(f);
		return false;	
	}
	fclose(f);
	return true;
}

bool load_cl(const std::string&file_name, ProgramState&old_state){
	FILE*f = fopen(file_name.c_str(), "r");

	if(!f){
		return false;
	}
	try{
		char buf_[20];
		ProgramState state;

		skip_string(f, file_tag);

		
		skip_string(f, background_tag);
		state.background_filename = read_string(f);
		fgets(buf_, sizeof(buf_)-1, f);


		skip_string(f, font_tag);
		state.font_desc = read_string(f);
		fgets(buf_, sizeof(buf_)-1, f);


		state.instance = new Instance;
		

		skip_string(f, instance_tag);
		fgets(buf_, sizeof(buf_)-1, f);
	
		string buf = buf_;
		if(buf.substr(buf.find_first_not_of(" "), 6) == "active")
			state.solution = new Solution;
		else
			state.solution = 0;
	
		
		
		for(;;){
			Label l;
			if(fscanf(f, "%lf%lf%lf%lf%lf%lf", &l.pos.x, &l.pos.y, &l.right, &l.top, &l.left, &l.bottom) != 6){
				break;
			}
			l.text= read_string(f);
			state.instance->add_label(l);
			if(state.solution)
				state.solution->active.push_back(read_angle_set(f));
		}
	
		int c = 0;
		while(c != EOF){
			c = fgetc(f);
			if(!isspace(c) && c != EOF)
				return false;
		}			

		delete old_state.instance;
		delete old_state.solution;		
		old_state = state;
	}catch(IllegalFileFormatError){
		fclose(f);
		return false;	
	}
	fclose(f);
	return true;
}
/*
void load(const char*file_name, Instance&instance, Solution&solution){
	FILE*f = fopen(file_name, "r");
	skip_string(f, instance_file_tag);
	soft_skip_string(f, instance_header);

	Instance new_instance;
	vector<AngleSet>active;
	for(;;){
		Label l;
		if(fscanf(f, "%lf%lf%lf%lf%lf%lf", &l.pos.x, &l.pos.y, &l.right, &l.top, &l.left, &l.bottom) != 6){
			instance = new_instance;
			solution.active = active;
			return;
		}
		l.text= read_string(f);
		new_instance.add_label(l);
		active.push_back(read_angle_set(f));
	}
}
*/
