#include "mip_model.h"

GRBEnv MIPModel::set_gurobi_env(const AlgorithmInfo &algo, bool debug) {
    GRBEnv gurobi_env = GRBEnv();

    if (debug)
        gurobi_env.set(GRB_IntParam_OutputFlag, 1);
    else
        gurobi_env.set(GRB_IntParam_OutputFlag, 0);

    if (algo.time_limit != std::numeric_limits<double>::min())
        gurobi_env.set(GRB_DoubleParam_TimeLimit, algo.time_limit);

    gurobi_env.set(GRB_IntParam_Threads, NMB_THREADS);

    return gurobi_env;
}



void MIPModel::MIP1::add_fixed_intervals_from_solution(Solution &sol, GRBModel& gurobi_model) {
    for (unsigned int i = 0; i < inst->count_labels(); ++i) {
        AngleSet a = sol.get_active(i);

        for (unsigned int j = 0; j < a.events.size(); j += 2) {
            int l_index = closest_to(a.events[j], interval_borders);
            int r_index = closest_to(a.events[j + 1], interval_borders);

            if (l_index == r_index)
                continue;

            for (int k = l_index; k < r_index; k++)
                gurobi_model.addConstr(x_i[i][k] == 1.0);
        }
    }
}


void MIPModel::MIP1::add_hard_conflict_constraints(GRBModel& gurobi_model) {

    for (unsigned int i = 0; i < inst->count_labels(); ++i) {

        /* handling label - label conflicts */
        const AngleSet &hc = inst->get_hard_conflicts(i);
        int nmb_conflicts = hc.events.size();

        for (int k = 0; k <= nmb_conflicts - 2; k += 2) {
            int start_index = closest_to(hc.events[k], interval_borders);
            int end_index = closest_to(hc.events[k + 1], interval_borders);

            for (int m = start_index; m < end_index; m++) {
                try {
                    // ensure that label cannot be active!
                    gurobi_model.addConstr(x_i[i][m] == 0.0, inst->get_label(i).text +
                                           " # " + to_string(i) +
                                           "_" + to_string(m));

                } catch (GRBException e) {
                    cout << "Gurobi Error Message:" << e.getMessage() << endl;
                    cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
                }
            }
        }

        /* handling label - dummy conflicts */
        for (unsigned j = 0; j < dummy_labels.size(); ++j) {

            AngleSet hc = calc_hard_conflicts(inst->get_label(i), dummy_labels[j]);

            for (int k = 0; k <= ((int)hc.events.size()) - 2; k += 2) {
                int start_index = closest_to(hc.events[k], interval_borders);
                int end_index = closest_to(hc.events[k + 1], interval_borders);

                for (int m = start_index; m < end_index; m++) {
                    try {
                        // ensure that label cannot be active!
                        gurobi_model.addConstr(x_i[i][m] == 0.0);
                    } catch (GRBException e) {
                        cout << "Gurobi Error Message:" << e.getMessage() << endl;
                        cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
                    }
                }
            }
        }
    }
}


void MIPModel::MIP1::add_conflict_constraints(GRBModel& gurobi_model) {


    const std::vector<conflict_information> &conflicts = inst->get_all_conflicts();

    for (const conflict_information &conflict : conflicts) {
        int nmb_conflict_borders = conflict.conflict_angles.events.size();

        for (int k = 0; k < nmb_conflict_borders; k += 2) {
            int start_index =
                    closest_to(conflict.conflict_angles.events[k], interval_borders);
            int end_index =
                    closest_to(conflict.conflict_angles.events[k + 1], interval_borders);

            for (int m = start_index; m < end_index; m++) {

                gurobi_model.addConstr(
                            x_i[conflict.label_id1][m] + x_i[conflict.label_id2][m] <= 1.0);
            }
        }
    }
}


void MIPModel::MIP1::add_monotonicity_constraints(GRBModel& gurobi_model) {
    int nmb_intervals = interval_borders.size() - 1;
    double max_nmb_intervals = algo->max_nmb_intervals;

    GRBLinExpr *bis = new GRBLinExpr[inst->count_labels()];
    char *senses = new char[inst->count_labels()];
    double *rhsValues = new double[inst->count_labels()];

    char *senses2 = new char[interval_borders.size()];
    double *rhsValues2 = new double[interval_borders.size()];
    for (unsigned int i = 0; i < interval_borders.size() - 1; i++) {
        senses2[i] = GRB_LESS_EQUAL;
        rhsValues2[i] = 0.0;
    }

    for (unsigned i = 0; i < inst->count_labels(); ++i) {
        senses[i] = GRB_LESS_EQUAL;
        rhsValues[i] = max_nmb_intervals;
    }

    for (unsigned i = 0; i < inst->count_labels(); ++i) {
        GRBLinExpr *mon = new GRBLinExpr[interval_borders.size() - 1];

        for (int j = 0; j < nmb_intervals; ++j) {
            mon[j] = x_i[i][j] - b_i[i][j] - x_i[i][(j + 1) % (nmb_intervals)];
            bis[i] += b_i[i][j];
        }
        if (interval_borders.size() <= 3) {
            // for up to three interval borders there is no need for monotonicity
            continue;
        }

        try {
            gurobi_model.addConstrs(mon, senses2, rhsValues2, NULL, nmb_intervals);
        } catch (GRBException e) {
            cout << "Gurobi Error Message:" << e.getMessage() << endl;
            cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
        }

        delete[] mon;
    }
    gurobi_model.addConstrs(bis, senses, rhsValues, NULL,
                            inst->count_labels()); // b_i
    delete[] bis;
    delete[] senses;
    delete[] rhsValues;
    delete[] senses2;
    delete[] rhsValues2;
}


void MIPModel::MIP1::compute_interval_borders(Solution& sol, bool use_initial_solution_force) {
    interval_borders.push_back(0.0);
    interval_borders.push_back(2 * M_PI);

    /* get interval borders from all soft conflicts */
    const std::vector<conflict_information> &conflicts = inst->get_all_conflicts();
    for (const conflict_information &conflict : conflicts) {
        interval_borders.insert(interval_borders.end(),
                                conflict.conflict_angles.events.begin(),
                                conflict.conflict_angles.events.end());
    }

    if (algo->use_hard_conflicts) {
        for (unsigned int i = 0; i < inst->count_labels(); ++i) {
            interval_borders.insert(interval_borders.end(),
                                    inst->get_hard_conflicts(i).events.begin(),
                                    inst->get_hard_conflicts(i).events.end());
        }
    }

    /* remove duplicates in 'points' */
    if (algo->use_hard_conflicts && points.size() > 0) {
        for (unsigned i = 0; i < inst->count_labels(); ++i) {
            Vector pos = inst->get_label(i).pos;

            for (int j = static_cast<int>(points.size()) - 1; j >= 0; --j) {
                if (pos == points[j]) {
                    points.erase(points.begin() + j);
                    break;
                }
            }
        }

        // add dummy labels from points (only necessary for hard conflicts)
        for (unsigned int i = 0; i < points.size(); ++i) {
            dummy_labels.emplace_back(points[i], 0, 0, 0, 0);
        }

        // add additional hard conflicts
        for (unsigned int i = 0; i < inst->count_labels(); ++i) {
            for (unsigned int j = 0; j < points.size(); ++j) {
                AngleSet hcAb = calc_hard_conflicts(inst->get_label(i), dummy_labels[j]);
                interval_borders.insert(interval_borders.end(), hcAb.events.begin(),
                                        hcAb.events.end());
            }
        }
    }

    // Initiallösung als Zwang verwenden
    if (use_initial_solution_force) {
        for (unsigned int i = 0; i < inst->count_labels(); ++i) {
            AngleSet a = sol.get_active(i);
            interval_borders.insert(interval_borders.end(), a.events.begin(),
                                    a.events.end());
        }
    }

    // sortieren der intervalgrenzen
    std::sort(interval_borders.begin(), interval_borders.end());

    // Duplikate filtern
    interval_borders.erase(
                std::unique(interval_borders.begin(), interval_borders.end()),
                interval_borders.end());
}


void MIPModel::MIP1::add_gurobi_vars(GRBModel& gurobi_model) {

    x_i.clear();
    b_i.clear();

    x_i.resize(inst->count_labels());
    b_i.resize(inst->count_labels());

    /* find smallest interval */
    double smallest_interval_length = 2 * M_PI;
    int nmb_intervals = interval_borders.size() - 1;
    for (int i = 0; i < nmb_intervals; i++) {
        smallest_interval_length =
                std::min(smallest_interval_length,
                         interval_borders[i + 1] - interval_borders[i]);
    }



    double *lower_bounds = new double[nmb_intervals]; // = 0.0
    double *upper_bounds = new double[nmb_intervals]; // = 1.0
    char *types = new char[nmb_intervals]; // GRB_BINARY
    double *interval_lengths = new double[nmb_intervals];
    double *bi_objval = new double[nmb_intervals];


    for (int i = 0; i < nmb_intervals; ++i) {
        lower_bounds[i] = 0.0;
        upper_bounds[i] = 1.0;
        types[i] = GRB_BINARY;
        interval_lengths[i] = interval_borders[i + 1] - interval_borders[i];
        bi_objval[i] = 0;
    }


    string *name_x = nullptr;
    string *name_b = nullptr;
    for (unsigned int i = 0; i < inst->count_labels(); i++) {

        GRBVar *vars_x =
                gurobi_model.addVars(lower_bounds, upper_bounds, interval_lengths,
                                     types, name_x, interval_borders.size() - 1);
        GRBVar *vars_b =
                gurobi_model.addVars(lower_bounds, upper_bounds, bi_objval, types,
                                     name_b, interval_borders.size() - 1);

        x_i[i].assign(vars_x, vars_x + interval_borders.size() - 1);
        b_i[i].assign(vars_b, vars_b + interval_borders.size() - 1);

        delete[] vars_x;
        delete[] vars_b;
    }


    delete[] lower_bounds;
    delete[] upper_bounds;
    delete[] name_x;
    delete[] name_b;
    delete[] types;
    delete[] interval_lengths;
    delete[] bi_objval;
}


int MIPModel::closest_to(const double number, const std::vector<double> &arr) {
    if (arr.size() == 0)
        return -1;

    double difference = abs(arr[0] - number);
    int index = 0;

    for (unsigned int i = 1; i < arr.size(); ++i) {
        if (difference > abs(arr[i] - number)) {
            difference = abs(arr[i] - number);
            index = i;
        }
    }

    return index;
}
