
/*
 * mip_model.h
 *
 *  Created on: 15.03.2011
 *      Author: andreas
 */

#ifndef MIP_MODEL_H_
#define MIP_MODEL_H_

#include "algorithms.hpp"
#include "instance.hpp"
#include "solution.hpp"
#include "collision.hpp"
#include "tools.hpp"

#include <algorithm>
#include <iomanip>
#include <utility>
#include <map>

#include "gurobi_c++.h"
#include <sstream>
#include <fstream>
using namespace std;

typedef pair<int, int> interval_border;
typedef pair<double, double> interval_border_values;

#define DEBUG_LINE cout << __PRETTY_FUNCTION__ << ", " << __LINE__ << endl;

#define NO_TIME_LIMIT std::numeric_limits<double>::max()
#define NMB_THREADS 1
#define ILP_OUTPUT false
#define TIME_LIMIT ONE_HOUR
#define ILP_WRITE_MODEL false
#define USE_CALLBACK true


namespace MIPModel{

int closest_to(const double number, const std::vector<double> &arr);

GRBEnv set_gurobi_env(const AlgorithmInfo &algo, bool debug = false);

    
class MIP1 {
public:

    bool run(const AlgorithmInfo &algorithm, const Instance &instance,
                           Solution &sol, bool use_initial_solution = false,
                           bool use_fixed_solution = false,
                           std::vector<Vector> points = {}) {

      /*
       * 1. Intervalle berechnen
       * 2. Zuweisung von Label-Konfliktpaaren zu intervallen bestimmen
       * 3. MIP Model vorbereiten [variable]
       * 4. MIP Constraints
       * 5. Lösen
       * 6. Lösung zurückübertragen
       */

      bool debug = ILP_OUTPUT;

      if (debug)
        std::cout << "debug ouput enabled." << endl;

      if (debug) {
          std::cout << std::boolalpha;
        std::cout << "max nmb intervals : " << algorithm.max_nmb_intervals << std::endl;
        std::cout << "use hard conflicts: " << algorithm.use_hard_conflicts << std::endl;
        std::cout << "use ini. sol.     : " << use_initial_solution << std::endl;
        std::cout << "use ini. sol. fo. : " << use_fixed_solution << std::endl;
        std::cout << "point size        : " << points.size() << std::endl;
        std::cout << "time limit        : " << algorithm.time_limit << " "
                  << ((algorithm.time_limit == std::numeric_limits<double>::min())
                          ? " min "
                          : " ") << std::endl;
        std::cout << std::noboolalpha;
      }



      if (instance.count_labels() == 0) {
        if (debug)
          cout << "instance contains 0 labels; skipping ILP" << endl;
        sol.optimal = true;
        sol.gap = 0.0;
        sol.obj_value = 0;
        sol.obj_bound = 0;
        return true;
      } else if (instance.count_labels() == 1) {
        if (debug)
          cout << "instance contains 1 label; skipping ILP" << endl;
        sol.set_active(0, AngleSet::all);
        sol.optimal = true;
        sol.gap = 0.0;
        sol.obj_value = 2 * M_PI;
        sol.obj_bound = 2 * M_PI;
        return true;
      }

      this->inst = &instance;
      this->algo = &algorithm;
      this->points = points;



      if (debug)
        cout << "number of labels: " << instance.count_labels() << endl;

      compute_interval_borders(sol, use_fixed_solution);

      int nmb_intervals = interval_borders.size() - 1;
      double max_nmb_intervals = algo->max_nmb_intervals;

      if (max_nmb_intervals < 0)
        max_nmb_intervals = interval_borders.size();



      // 2. Gurobi init stuff
      GRBEnv gurobi_env = set_gurobi_env(*algo, debug);
      GRBModel gurobi_model = GRBModel(gurobi_env);
      gurobi_model.set(GRB_IntAttr_ModelSense, -1);



      // Gurobi variables
      if (debug)
        std::cout << "adding variables..." << std::endl;


      add_gurobi_vars(gurobi_model);


      if (debug)
        std::cout << "adding variables... done!" << std::endl;

      try {
        gurobi_model.update();
      } catch (GRBException e) {
        cout << "Gurobi Error Message:" << e.getMessage() << endl;
        cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
      }


      /***** CONSTRAINTS *******/
      if (debug)
        std::cout << "adding monotonicity constraints.." << std::endl;

      add_monotonicity_constraints(gurobi_model);

      if (debug)
        std::cout << "adding monotonicity constraints...done!" << std::endl;

      if (debug)
        std::cout << "adding conflict constraints..." << std::endl;

      add_conflict_constraints(gurobi_model);

      if (debug)
        std::cout << "adding conflict constraints...done!" << std::endl;

      if (algo->use_hard_conflicts) {
        if (debug)
          std::cout << "adding hard conflict constraints..." << std::endl;

        add_hard_conflict_constraints(gurobi_model);

        if (debug)
          std::cout << "adding hard conflict constraints...done!" << std::endl;
      }

      // Initiallösung als verwenden
      if (use_fixed_solution) {
        add_fixed_intervals_from_solution(sol, gurobi_model);
      }

      if (debug)
        std::cout << "starting optimization..." << std::endl;

      try {
        gurobi_model.optimize();
      } catch (GRBException e) {
        cout << "Gurobi Error Message:" << e.getMessage() << endl;
        cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
      }

      int gurobi_status = gurobi_model.get(GRB_IntAttr_Status);

      if (gurobi_status == GRB_OPTIMAL || gurobi_status == GRB_INTERRUPTED ||
          gurobi_status == GRB_TIME_LIMIT) {
        // 6. Zurückübersetzen
        if (debug)
          std::cout << "optimal solution found!" << std::endl;

        for (unsigned i = 0; i < inst->count_labels(); i++) {
          bool has_been_active = false;
          bool active = false;
          AngleSet sol_i;
          double start_angle = -1;
          double end_angle = -1;

          for (int j = 0; j < nmb_intervals; j++) {
            try {
              active = (x_i[i][j].get(GRB_DoubleAttr_X) == 1);
            } catch (GRBException e) {
              std::cout << i << " " << j << std::endl;
              cout << "Gurobi Error Message:" << e.getMessage() << endl;
              cout << "Gurobi Error Code:" << e.getErrorCode() << endl;
            }

            if (active && !has_been_active) {
              start_angle = interval_borders[j];
            } else if (!active && has_been_active) {
              sol_i = sol_i | AngleSet(start_angle, interval_borders[j]);
              end_angle = -1;
              start_angle = -1;
            }
            has_been_active = active;
          }

          if (end_angle == -1 && start_angle != -1) {
            sol_i = sol_i | AngleSet(start_angle, 2 * M_PI);
          }

          sol.set_active(i, sol_i);
        }

        //std::cout << "sol value: " << sol.active_time() << std::endl;

        if (gurobi_status == GRB_OPTIMAL)
          return true;
        else {
          double obj_val = gurobi_model.get(GRB_DoubleAttr_ObjVal);
          double obj_bound = gurobi_model.get(GRB_DoubleAttr_ObjBound);
          sol.gap = 1 - (obj_val / obj_bound);
          sol.obj_value = obj_val;
          sol.obj_bound = obj_bound;
          return false;
        }
      }

      assert(false);
      return false;
    }

private:
    const Instance* inst = nullptr;
    const AlgorithmInfo* algo = nullptr;
    std::vector<Vector> points;

    std::vector<std::vector<GRBVar>> x_i;
    std::vector<std::vector<GRBVar>> b_i;
    std::vector<double> interval_borders;
    std::vector<Label> dummy_labels;


void compute_interval_borders(Solution& sol, bool use_initial_solution_force = false);

void add_gurobi_vars(GRBModel& gurobi_model);
void add_monotonicity_constraints(GRBModel& gurobi_model);
void add_conflict_constraints(GRBModel& gurobi_model);
void add_hard_conflict_constraints(GRBModel& gurobi_model);
void add_fixed_intervals_from_solution(Solution &sol, GRBModel& gurobi_model);

};

};
#endif /* MIP_MODEL_H_ */
