#ifndef MIP_MODEL2_H
#define MIP_MODEL2_H

namespace MIPModel {

struct ConflictBinaryVars {
  GRBVar x1, x2, x;
  GRBVar k1, k2, k3, k4;
  GRBVar y1, y2;

  GRBVar z;
  GRBVar l1, l2, l3, l4;

  GRBVar active;
};



bool calcSolutionByMip2(const AlgorithmInfo &algo, const Instance &inst, Solution &sol, bool use_initial_solution = false) {

    bool debug = ILP_OUTPUT;

    if (inst.count_labels() == 0) {
        if (debug)
            cout << "instance contains 0 labels; skipping ILP" << endl;
        sol.optimal = true;
        sol.gap = 0.0;
        sol.obj_value = 0;
        sol.obj_bound = 0;
        return true;
    }

    // DEBUG
    assert(algo.max_nmb_intervals ==
           1); // algorithm does not support multiple intervals

    DEBUG_LINE
            if (debug)
            cout << "number of labels: " << inst.count_labels() << endl;
    DEBUG_LINE

            // Gurobi
            DEBUG_LINE;
    GRBEnv gurobi_env = set_gurobi_env(algo, true);
    GRBModel model = GRBModel(gurobi_env);
    model.set(GRB_IntAttr_ModelSense, -1);


    DEBUG_LINE

            // Variablen
            vector<GRBVar> a(inst.count_labels());
    vector<GRBVar> b(inst.count_labels());
    vector<GRBVar> active(inst.count_labels());
    vector<vector<vector<ConflictBinaryVars>>> conflicts(inst.count_labels());
    vector<GRBVar> hcbvars;
    int hcbvars_counter = 0;

    for (unsigned int i = 0; i < inst.count_labels(); ++i) {
        conflicts[i].resize(inst.count_labels());
    }

    for (unsigned int i = 0; i < inst.count_labels(); ++i) {
        a[i] = model.addVar(0.0, 2 * M_PI, -1.0, GRB_CONTINUOUS,
                            "a" + std::to_string(i));
        b[i] = model.addVar(
                    0.0, 4 * M_PI, 1.0, GRB_CONTINUOUS,
                    "b" + std::to_string(i)); // objective function is sum b[i] - a[i]

        for (unsigned int j = i + 1; j < inst.count_labels(); ++j) {
            AngleSet collisions =
                    calc_collision(inst.get_label(i), inst.get_label(j));

            if (collisions.events.size() == 0)
                continue;

            conflicts[i][j].resize(collisions.events.size() / 2);
            for (unsigned int k = 0; k < collisions.events.size() / 2; ++k) {
                string identifier = "[" + to_string(i) + "][" + to_string(j) + "][" +
                        to_string(k) + "]";
                conflicts[i][j][k].x1 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "x1" + identifier);
                conflicts[i][j][k].x2 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "x2" + identifier);
                conflicts[i][j][k].x =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "x" + identifier);

                conflicts[i][j][k].k1 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "k1" + identifier);
                conflicts[i][j][k].k2 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "k2" + identifier);
                conflicts[i][j][k].k3 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "k3" + identifier);
                conflicts[i][j][k].k4 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "k4" + identifier);

                conflicts[i][j][k].y1 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "y1" + identifier);
                conflicts[i][j][k].y2 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "y2" + identifier);

                conflicts[i][j][k].l1 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "l1" + identifier);
                conflicts[i][j][k].l2 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "l2" + identifier);
                conflicts[i][j][k].l3 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "l3" + identifier);
                conflicts[i][j][k].l4 =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "l4" + identifier);

                conflicts[i][j][k].z =
                        model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "z" + identifier);
            }

            if (algo.use_hard_conflicts) {

                AngleSet hcAb =
                        calc_hard_conflicts(inst.get_label(i), inst.get_label(j));
                AngleSet hcaB =
                        calc_hard_conflicts(inst.get_label(j), inst.get_label(i));

                for (unsigned int l = 0;
                     l < (hcAb.events.size() + hcaB.events.size()) / 2; ++l) {
                    hcbvars.push_back(model.addVar(0.0, 1.0, 0.0, GRB_BINARY,
                                                   "hcbvar" + to_string(hcbvars.size())));
                }
            }
        }
    }
    // Variablen dem model hinzufügen
    model.update();

    const double M = 13; // > 4 * pi

    // Constraints
    for (unsigned int i = 0; i < inst.count_labels(); ++i) {
        model.addConstr(a[i] <= b[i]);
        model.addConstr(b[i] - a[i] <= 2 * M_PI);

        for (unsigned int j = i + 1; j < inst.count_labels(); ++j) {
            AngleSet collisions =
                    calc_collision(inst.get_label(i), inst.get_label(j));

            if (collisions.events.size() == 0)
                continue;

            for (unsigned int z = 0; z < collisions.events.size(); z += 2) {
                double left_border = collisions.events[z];
                double right_border = collisions.events[z + 1];

                ConflictBinaryVars &bvars = conflicts[i][j][(z / 2)];

                // bvars.x = 1, wenn Label i Konfliktregion uberschneidet
                model.addConstr(bvars.x1 >= (right_border - a[i]) / M);
                model.addConstr(bvars.x2 >= (b[i] - left_border) / M);
                model.addConstr(bvars.x >= bvars.x1 + bvars.x2 - 1);

                // Wenn bvars.x = 1 dann muss Konflikt vermieden werden
                model.addConstr(bvars.k1 + bvars.k2 + bvars.k3 + bvars.k4 >= bvars.x);

                // Konfliktvermeidung
                model.addConstr(b[j] <= left_border + (1 - bvars.k4) * M);
                model.addConstr(b[j] <= a[i] + (1 - bvars.k3) * M);

                model.addConstr(a[j] >= b[i] - (1 - bvars.k1) * M);
                model.addConstr(a[j] >= right_border - (1 - bvars.k2) * M);
                // Zusatz wegen unrolling
                model.addConstr(bvars.y1 + bvars.y2 >= bvars.k1 + bvars.k2);
                model.addConstr(b[j] <= a[i] + 2 * M_PI + (1 - bvars.y1) * M);
                model.addConstr(b[j] <= left_border + 2 * M_PI + (1 - bvars.y2) * M);

                //
                model.addConstr(bvars.z >= (b[i] - (left_border + 2 * M_PI)) / M);

                model.addConstr(bvars.l1 + bvars.l2 + bvars.l3 + bvars.l4 >= bvars.z);

                // Konfliktvermeidung
                model.addConstr(b[j] <= left_border + (1 - bvars.l1) * M,
                                "b[" + to_string(j) + "] <= " + to_string(left_border));
                model.addConstr(a[j] >= b[i] - 2 * M_PI - (1 - bvars.l2) * M,
                                "a[" + to_string(j) + "] >= b[i] - 2pi");
                model.addConstr(a[j] >= right_border - (1 - bvars.l3) * M,
                                "a[" + to_string(j) + "] >= " +
                                to_string(right_border));

                model.addConstr(a[i] >= right_border - (1 - bvars.l2) * M);
                model.addConstr(b[j] <= left_border + 2 * M_PI +
                                (1 - (bvars.l2 + bvars.l3)) * M);

                model.addConstr(a[j] >= b[i] - 2 * M_PI - (1 - bvars.l4) * M);
                model.addConstr(b[j] <= a[i] + (1 - bvars.l4) * M);
            }

            if (algo.use_hard_conflicts) {
                AngleSet hcAb =
                        calc_hard_conflicts(inst.get_label(i), inst.get_label(j));
                AngleSet hcaB =
                        calc_hard_conflicts(inst.get_label(j), inst.get_label(i));

                for (unsigned int l = 0; l < hcAb.events.size(); l += 2) {
                    double left = hcAb.events[l];
                    double right = hcAb.events[l + 1];

                    model.addConstr(a[i] >= right - (1 - hcbvars[hcbvars_counter]) * M);
                    model.addConstr(b[i] <= left + hcbvars[hcbvars_counter] * M);
                    model.addConstr(b[i] <= left + 2 * M_PI);

                    hcbvars_counter++;
                }

                for (unsigned int l = 0; l < hcaB.events.size(); l += 2) {
                    double left = hcaB.events[l];
                    double right = hcaB.events[l + 1];

                    model.addConstr(a[j] >= right - (1 - hcbvars[hcbvars_counter]) * M);
                    model.addConstr(b[j] <= left + hcbvars[hcbvars_counter] * M);
                    model.addConstr(b[j] <= left + 2 * M_PI);

                    hcbvars_counter++;
                }
            }
        }
    }

    // Initiallösung verwenden
    if (use_initial_solution) {

        for (unsigned int i = 0; i < inst.count_labels(); ++i) {
            AngleSet angleset = sol.get_active(i);

            if (angleset.events.size() == 2) {
                a[i].set(GRB_DoubleAttr_Start, angleset.events[0]);
                b[i].set(GRB_DoubleAttr_Start, angleset.events[1]);
            } else if (angleset.events.size() == 4) {
                a[i].set(GRB_DoubleAttr_Start, angleset.events[3]);
                b[i].set(GRB_DoubleAttr_Start, angleset.events[1] + 2 * M_PI);
            }
        }
    }
    // Initiallösung verwenden

    model.optimize();
    if (debug)
        std::cout << "starting optimization...done!" << std::endl;

    /* model.computeIIS(); */
    /* model.write("debug.ilp"); */

    if (model.get(GRB_IntAttr_Status) == GRB_OPTIMAL) {
        // 6. Zurückübersetzen
        if (debug)
            std::cout << "optimal solution found!" << std::endl;

        for (unsigned i = 0; i < inst.count_labels(); i++) {
            AngleSet sol_i = AngleSet::empty;

            double ai = a[i].get(GRB_DoubleAttr_X);
            double bi = b[i].get(GRB_DoubleAttr_X);

            double epsilon = 0.0000001;
            if (abs(ai - bi) <= epsilon) {
                sol.set_active(i, AngleSet::empty);
                continue;
            }

            if (bi > 2 * M_PI) {

                if (ai < 2 * M_PI)
                    sol_i = AngleSet(0, bi - 2 * M_PI) | AngleSet(ai, 2 * M_PI);
                else
                    sol_i = AngleSet(0, bi - 2 * M_PI);

            } else {
                sol_i = AngleSet(ai, bi);
            }

            sol.set_active(i, sol_i);
        }
        return true;
    }

    if (model.get(GRB_IntAttr_Status) == GRB_TIME_LIMIT) {
        // 6. Zurückübersetzen
        if (debug)
            std::cout << "time limit!" << std::endl;

        for (unsigned i = 0; i < inst.count_labels(); i++) {
            AngleSet sol_i = AngleSet::empty;

            double ai = a[i].get(GRB_DoubleAttr_Xn);
            double bi = b[i].get(GRB_DoubleAttr_Xn);

            double epsilon = 0.0000001;
            if (abs(ai - bi) <= epsilon) {
                sol.set_active(i, AngleSet::empty);
                continue;
            }

            if (bi > 2 * M_PI) {

                if (ai < 2 * M_PI)
                    sol_i = AngleSet(0, bi - 2 * M_PI) | AngleSet(ai, 2 * M_PI);
                else
                    sol_i = AngleSet(0, bi - 2 * M_PI);

            } else {
                sol_i = AngleSet(ai, bi);
            }

            sol.set_active(i, sol_i);
        }
        // determine gap
        double obj_val = model.get(GRB_DoubleAttr_ObjVal);
        double obj_bound = model.get(GRB_DoubleAttr_ObjBound);
        sol.gap = 1 - (obj_val / obj_bound);
        sol.obj_value = obj_val;
        sol.obj_bound = obj_bound;

        return false;
    }
    return false;
}

}


#endif // MIP_MODEL2_H
