#pragma GCC diagnostic ignored "-Wwrite-strings" 

#include <cmath>
#include <algorithm>
#include <cstring>
#include <iostream>
#include <unordered_map>

#include "angle_set.hpp"
#include "instance.hpp"
#include "solution.hpp"
#include "collision.hpp"

#include "mip_model.h"
#include "mip_model2.h"
#include "mip_model_no_flicker.h"
#include "greedy.h"
#include "Unionfind.h"
#include "tools.hpp"


Solution calc_mip_solution_no_flickering(Instance& inst, bool use_hard_conflicts){
  Solution sol(inst);
  Timer t;
  AlgorithmInfo algo;
  algo.name = AlgorithmInfo::MIP_NOFLICKER;
  algo.use_hard_conflicts = use_hard_conflicts;
  inst.compute_conflicts();
  MIPModel::calcSolutionByMipNoFlickering(algo, inst, sol);
  cout << "time: " << t.elapsed() << "ms" << endl;
  return sol;
}


void compute_connected_components(const Instance& inst, std::vector<Instance>& instances, std::vector<std::vector<int>>& retranslate_info) {
    int nmb_labels = inst.count_labels();

    Unionfind uf(nmb_labels);
    /* iterate over all conflicts */
    const std::vector<conflict_information>& conflicts = inst.get_all_conflicts();
    for (const conflict_information& conflict: conflicts)
        uf.merge(conflict.label_id1, conflict.label_id2);

    std::unordered_map<int, int> mapping;
    /* iterate over all labels */
    for (int i = 0; i < nmb_labels; ++i) {
        int value = uf.find(i);
        const auto pos = mapping.find(value);

        if ( pos == mapping.end()) {
            retranslate_info.push_back(std::vector<int>{i});
            mapping[value] = retranslate_info.size() - 1;
            instances.emplace_back(Instance());
            (instances.back()).add_label(inst.get_label(i));
        } else {
            retranslate_info[pos->second].push_back(i);
            instances[pos->second].add_label(inst.get_label(i));
        }
    }
}

Solution calc_mip_solution_fast(AlgorithmInfo algo, Instance& inst, bool use_initial_solution = false) {

    inst.compute_conflicts();
    vector< vector < int > > retranslate_info;
    vector <Solution> solutions;
    vector <Instance> instances;

    compute_connected_components(inst, instances, retranslate_info);
    for (const Instance& small_inst: instances)
        solutions.emplace_back(small_inst);

    bool guaranteed_optimal_solution = true;
    int largest_instance = instances[0].count_labels();
    const int large_instance = 20;
    
    for (unsigned int i = 0; i < instances.size(); i++){
        Instance& small_instance = instances[i];
        Solution& small_solution = solutions[i];
        
        largest_instance = std::max(static_cast<int>(instances[i].count_labels()), largest_instance);
        
        if (instances[i].count_labels() > large_instance)
            cout << "computing solution for instance with " << instances[i].count_labels() << " vertices..." << endl;
        

        if (use_initial_solution) {
            AlgorithmInfo algo_greedy(algo);
            algo_greedy.name = AlgorithmInfo::GreedyMax;
            Greedy g(algo_greedy, small_instance, small_solution);
        }
        
        small_instance.compute_conflicts();
        MIPModel::MIP1 m;
        switch (algo.name) {
            case AlgorithmInfo::MIP:
                guaranteed_optimal_solution &= m.run(algo, small_instance, small_solution, use_initial_solution);
                break;
            case AlgorithmInfo::MIP2:
                guaranteed_optimal_solution &= MIPModel::calcSolutionByMip2(algo, small_instance, small_solution, use_initial_solution);
                break;
            default:
                assert(false);
        }


        if (instances[i].count_labels() > large_instance)
            cout << " .. done!" << endl;
    }
    std::cout << "largest instance: " << largest_instance << ", nmb connected components:" << instances.size() << std::endl;
    
    
    Solution sol(inst);
    
    //compute gap
    double total_obj_value = 0.0;
    double total_obj_bound = 0.0;
    for (unsigned int i = 0; i < solutions.size(); i++){
        total_obj_value += solutions[i].obj_value;
        total_obj_bound += solutions[i].obj_bound;
    }
    double gap = 0.0;
    if (total_obj_bound > 0){
        gap = 1.0 - total_obj_value/total_obj_bound;
    }
    sol.gap = gap;
    
    //set solutions
    for (unsigned int i = 0; i < instances.size(); ++i){
        for (unsigned int j = 0; j < instances[i].count_labels(); ++j){
            int label_index = retranslate_info[i][j];
            AngleSet a = solutions[i].get_active(j);
            sol.set_active(label_index, a);
        }
    }
    
    sol.optimal = guaranteed_optimal_solution;
    
    return sol;
}

Solution calc_greedy(Instance& inst, const AlgorithmInfo& algo) {
    inst.compute_conflicts();

    Solution sol(inst);
    Greedy g(algo, inst, sol);

    sol.gap = 0;
    return sol;
}



Solution calc_appr(const Instance&inst, const AlgorithmInfo& algo_mip, int app = 4, std::vector<Solution>* solutions = nullptr)
{
    assert(algo_mip.name == AlgorithmInfo::MIP);

    int bcw = (int)sqrt(app);
    const bool debug = false;
	Solution ret_sol(inst);
    if (solutions)
        solutions->clear();

    std::vector<Vector> points = {};
    if (algo_mip.use_hard_conflicts) {
        for (unsigned int i = 0; i < inst.count_labels(); ++i) {
            points.push_back(inst.get_label(i).pos);
        }
    }



	//calculate grid size
	Label cur_label;
	double cellsize = 1, cur;
	double min = 0;
    for (unsigned int i = 0; i < inst.count_labels(); i++)	{
		cur_label = inst.get_label(i);
		if (cur_label.pos.x > min) min = cur_label.pos.x;
		if (cur_label.pos.y > min) min = cur_label.pos.y;

		if (cur_label.top > cur_label.bottom) {
			if(cur_label.left > cur_label.right) {
				cur = sqrt(cur_label.top*cur_label.top+cur_label.left*cur_label.left);
				if (cur > cellsize) cellsize = cur;
			}
			else {//top > bottom && left <= right
				cur = sqrt(cur_label.top*cur_label.top+cur_label.right*cur_label.right);
				if (cur > cellsize) cellsize = cur;
			}
		}
		else { //top <= bottom
			if(cur_label.left > cur_label.right) {
				cur = sqrt(cur_label.bottom*cur_label.bottom+cur_label.left*cur_label.left);
				if (cur > cellsize) cellsize = cur;
			}
			else { //top <= bottom && left <= right
				cur = sqrt(cur_label.bottom*cur_label.bottom+cur_label.right*cur_label.right);
				if (cur > cellsize) cellsize = cur;
			}
		}
	}

	cellsize *= 2;
	double bigcellsize = cellsize * bcw;
	multimap<Vector, int> non_empty_cells;

	//fill map with lists of labels
	multimap<Vector, int>::iterator it,it_end;
	pair<multimap<Vector, int>::iterator, multimap<Vector, int>::iterator> inner_it;
	//Testausgabe
	//--

	it=non_empty_cells.begin();
	it_end=non_empty_cells.end();
	Vector cur_key = it->first;
	Instance small_inst;
	map<int, int> zuw;
	multimap<int, int>::iterator in_it, in_it_end;
	int a,b;

	min = 0;
	for (int k = 0; k < app; k++) {
		//int k = iteration;
		int num = 0;
		Solution sol(inst);
		a = k % bcw;
		b = k / bcw;
		non_empty_cells.clear();
        for (unsigned int i = 0; i < inst.count_labels(); i++)	{
			cur_label = inst.get_label(i);
			Vector temp_vector;
			temp_vector.x = floor((cur_label.pos.x+min-(a*cellsize))/bigcellsize);
			temp_vector.y = floor((cur_label.pos.y+min-(b*cellsize))/bigcellsize);

			non_empty_cells.insert(pair<Vector,int>(temp_vector, i));
		}

		it=non_empty_cells.begin();
		it_end=non_empty_cells.end();
		while(it!=it_end) {
			if (cur_key.x == it->first.x && cur_key.y == it->first.y) {
				int ax,ay;
				ax = (int)(floor((inst.get_label(it->second).pos.x+min)/cellsize));
				ay = (int)(floor((inst.get_label(it->second).pos.y+min)/cellsize));
				if( ax < 0) ax += -ax * bcw;
				if( ay < 0) ay += -ay * bcw;
				if(  ax % bcw == a || ay % bcw == b) {
				} else {
					int temp = small_inst.add_label(inst.get_label((it->second)));
					zuw[temp] = it->second;
				}
				it++;
			}
			else{ //new cell
				Solution small_sol(small_inst);

				if(zuw.size() > 0){
                    small_inst.compute_conflicts();
                    MIPModel::MIP1 m;
                    m.run(algo_mip, small_inst, small_sol, false, false, points);

                    num++;
					in_it = zuw.begin();
					in_it_end = zuw.end();
					while (in_it != in_it_end){
						sol.set_active(in_it->second, small_sol.get_active(in_it->first));
						in_it++;
					}
				}

				while(small_inst.count_labels() > 0) {
					small_inst.remove_label(0);
				}
				zuw.clear();

				cur_key = it->first;
			}
		}

		//else part again to calculate the last gridcell
        Solution small_sol(small_inst);
        small_inst.compute_conflicts();
        MIPModel::MIP1 m;
        m.run(algo_mip, small_inst, small_sol, false, false, points);
        num++;
		//}

		if(zuw.size() > 0){
            small_inst.compute_conflicts();
            MIPModel::MIP1 m;
            m.run(algo_mip, small_inst, small_sol, false, false, points);

            num++;
			in_it = zuw.begin();
			in_it_end = zuw.end();
			while (in_it != in_it_end){
				sol.set_active(in_it->second, small_sol.get_active(in_it->first));
				in_it++;
			}
		}

		while(small_inst.count_labels() > 0) {
			small_inst.remove_label(0);
		}
		zuw.clear();

		//--
		if(debug)std::cout << "------------------------------------"<<std::endl;
		if(debug)std::cout << sol.active_time() <<std::endl;

        if (solutions != nullptr)
            solutions->push_back(sol);

		if(sol.active_time() >= ret_sol.active_time()) 
			ret_sol = sol;

		if(debug)cout << "a: " << a << " b: " << b <<endl; 

	}
	if(debug)std::cout <<std::endl << "app: " << app << "bcw: " << bcw << std::endl;
	return ret_sol;
}

Solution calc_app(Instance& inst, int nmb_max_intervals, bool use_hard_conflicts, int app=4, AlgorithmInfo::AlgorithmName greedy_algo = AlgorithmInfo::UNSPECIFIED) {
    AlgorithmInfo algo;
    algo.name = AlgorithmInfo::MIP;
    algo.max_nmb_intervals = nmb_max_intervals;
    algo.use_hard_conflicts = use_hard_conflicts;

    std::vector<Solution> solutions;

    inst.compute_conflicts();
    Solution best_solution = calc_appr(inst, algo, app, &solutions);

    if (greedy_algo == AlgorithmInfo::UNSPECIFIED)
        return best_solution;

    /* run additionally a greedy algorithm */
    AlgorithmInfo algo_greedy(algo);
    algo_greedy.name = greedy_algo;

    for (Solution& sol: solutions) {
        Greedy g(algo_greedy, inst, sol, true);

        if (best_solution.active_time() < sol.active_time())
            best_solution = sol;
    }

    return best_solution;
}
