#pragma GCC diagnostic ignored "-Wwrite-strings"
#pragma GCC diagnostic ignored "-Wswitch"

#include "viewer.hpp"
#include "instance.hpp"
#include "solution.hpp"
#include "solution_algo.hpp"
#include "geo.hpp"
#include "io.hpp"
#include "collision.hpp"
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <cairo.h>
#include <cmath>
#include <sys/time.h>
#include <cstring>
#include <clocale>
#include <cstdio>
#include <string>
#include <unistd.h>


#include <sstream>
#include <iostream>

int screenshot_counter = 0;
bool b_take_screenshots = FALSE;
bool draw_label_rectangle = TRUE;



//$ linker: `pkg-config --libs cairo gtk+-2.0 pango`
//$ compiler: `pkg-config --cflags cairo gtk+-2.0 pango`
//$ compiler: -Wno-write-strings

using namespace std;

// -------------------------------------------------------- //

ProgramState program_state;// = {0, 0, ""};
string save_to_file = "";
cairo_surface_t*background = 0;
PangoFontDescription * font_desc = 0;
GtkWidget*drawing_area = 0, *window = 0;

vector<int>label_text_width, label_text_height;

// -------------------------------------------------------- //

inline void afftrans_to_cairo_matrix(AffTrans a, cairo_matrix_t*m){
	m->xx = a.m.a;
	m->xy = a.m.b;
	m->yx = a.m.c;
	m->yy = a.m.d;
	m->x0 = a.t.x;
	m->y0 = a.t.y;
}

// -------------------------------------------------------- //

//! Three 2D coordinate systems are used:
//!  * world space
//!  * map space
//!  * camera space
//!
//! world_trans transformed world space points into
//! map space points and camera_trans transforms points
//! from the map space into camera points.
//! 
//! All these affine transformations are positive 
//! isometries (i.e. determinant = +1).
//! 
//! The world space contains indexes the texels of the
//! background image and the rotation centers of the labels.
//!
//! The map space is the system in which the labels are placed.
//! 
//! The camera space are coordinates on the window.
AffTrans world_trans;
Matrix camera_trans;

static const unsigned no_label = (unsigned)-1;
unsigned selected_label = no_label;

bool draw_hidden_labels_transparent = true;
bool hide_solution_circle = false;
bool frame_limiter = true;
bool draw_camera_mid = false;

int background_x = 0;
int background_y = 0;

void camera_to_map(int&x, int&y){
	Vector t = geo_util::inv(camera_trans)*Vector(x, y);
	x = int(round(t.x));
	y = int(round(t.y));
}



// -------------------------------------------------------- //

void draw_background(cairo_t *cr){
	cairo_save(cr);
	if(background){
		int bg_w = cairo_image_surface_get_width(background);
		int bg_h = cairo_image_surface_get_height(background);

		cairo_matrix_t temp;
		afftrans_to_cairo_matrix(world_trans, &temp);
		cairo_transform(cr, &temp);
        
        //cairo_scale(cr, 1.5, 1.5);
    
		cairo_set_source_surface(cr, background, -bg_w/2 + background_x, -bg_h/2 + background_y);
		cairo_paint(cr);

		// no bounding rectangle but slower
		/*{
			cairo_rectangle(cr, -bg_w/2, -bg_h/2, bg_w, bg_h);
			cairo_set_source_surface(cr, background, -bg_w/2, -bg_h/2);
			cairo_fill(cr);
		}*/
	}else{
		cairo_set_source_rgb (cr, 1, 1, 1);
		cairo_paint(cr);
	}
	cairo_restore(cr);
}

void draw_label(cairo_t*cr, unsigned label_index, bool transparent, double rot_angle){
	cairo_save(cr);
	const Label&l = program_state.instance->get_label(label_index);

	Vector v = world_trans*l.pos;
	cairo_translate(cr, v.x, v.y);

	double alpha = 1.0;
	if(transparent)
		alpha = 0.2;

	if(label_index != selected_label)
		cairo_set_source_rgba (cr, 1, 1, 1, alpha);
	else
		cairo_set_source_rgba (cr, 0.7, 0.7, 1, alpha);

	if (draw_label_rectangle)
	{
		cairo_rectangle(cr, -l.left, -l.top, l.left + l.right, l.top + l.bottom);
		cairo_fill(cr);
	}

	cairo_set_source_rgba (cr, 1, 0, 0, alpha);

	const int s = 5;
	cairo_move_to(cr, -s, 0);
	cairo_line_to(cr, +s, 0);
	cairo_move_to(cr, 0, -s);
	cairo_line_to(cr, 0, +s);

	cairo_stroke(cr);

	cairo_set_line_width (cr, 2.0);
	if(program_state.solution && !hide_solution_circle){
		const AngleSet&angle_set = program_state.solution->get_active(label_index);

        for(unsigned int i=0; i<angle_set.events.size(); i+=2){
			cairo_arc (cr, 0, 0, 10.0, angle_set.events[i]-rot_angle, angle_set.events[i+1]-rot_angle);
			cairo_stroke (cr);
		}
	}

	cairo_set_source_rgba (cr, 0, 0, 0, alpha);

	if (draw_label_rectangle)
	{
		cairo_rectangle(cr, -l.left, -l.top, l.left + l.right, l.top + l.bottom);
		cairo_stroke(cr);
	}

	if(!font_desc){
		cairo_text_extents_t te;
		cairo_text_extents(cr, l.text.c_str(), &te);

		cairo_translate(cr, 
			-l.left + (l.left+l.right)/2 - te.width/2, 
			-l.top + (l.top+l.bottom)/2 + te.height/2);

		cairo_show_text (cr, l.text.c_str());

		label_text_width.resize(std::max((int)label_text_width.size(), (int)label_index+1));
		label_text_width[label_index] = te.width;
		label_text_height.resize(std::max((int)label_text_height.size(), (int)label_index+1));
		label_text_height[label_index] = te.height;
    } else {
		PangoLayout *layout = pango_cairo_create_layout (cr);
		pango_layout_set_text (layout, l.text.c_str(), -1);
		pango_layout_set_font_description (layout, font_desc);
		int text_width, text_height;
		pango_layout_get_size (layout, &text_width, &text_height);
		text_width /= PANGO_SCALE;
		text_height /= PANGO_SCALE;
		cairo_move_to(cr, 
			-l.left + (l.left+l.right)/2 - text_width/2, 
			-l.top + (l.top+l.bottom)/2 - text_height/2);
		pango_cairo_show_layout (cr, layout);
		g_object_unref(layout);

		label_text_width.resize(std::max((int)label_text_width.size(), (int)label_index+1));
		label_text_width[label_index] = text_width;
		label_text_height.resize(std::max((int)label_text_height.size(), (int)label_index+1));
		label_text_height[label_index] = text_height;
	}

	cairo_restore(cr);
}

/*
void draw_flight_target(cairo_t*cr){
	if(flight_target.x != 0.0 || flight_target.y != 0.0){
		cairo_save(cr);
		cairo_translate(cr, flight_target.x, flight_target.y);

		cairo_set_line_width (cr, 5.0);
			cairo_set_source_rgb (cr, 0, 1, 0);

		const int s = 20;
		cairo_move_to(cr, -s, 0);
		cairo_line_to(cr, +s, 0);
		cairo_move_to(cr, 0, -s);
		cairo_line_to(cr, 0, +s);

		cairo_stroke(cr);
		cairo_restore(cr);
	}
}
*/

void draw_all(cairo_t*cr, int w, int h){
	cairo_translate(cr, w/2, h/2);
	
	cairo_save(cr);
	cairo_matrix_t temp;
	afftrans_to_cairo_matrix(camera_trans, &temp);
	cairo_transform(cr, &temp);

	
	draw_background(cr);

	double rot_angle = geo_util::rot_angle(world_trans.m);

	if(!program_state.solution){
		for(unsigned i=0; i<program_state.instance->count_labels(); ++i)
			draw_label(cr, i, false, rot_angle);
	}else{
	  for(unsigned i=0; i<program_state.instance->count_labels(); ++i) {
			cairo_save(cr);
			const Label&l = program_state.instance->get_label(i);

			Vector v = world_trans*l.pos;
			cairo_translate(cr, v.x, v.y);

			cairo_set_source_rgba (cr, 1, 0, 0, 1);

			const int s = 5;
			cairo_move_to(cr, -s, 0);
			cairo_line_to(cr, +s, 0);
			cairo_move_to(cr, 0, -s);
			cairo_line_to(cr, 0, +s);

			cairo_stroke(cr);
			cairo_restore(cr);
		}

      if(draw_hidden_labels_transparent)
			for(unsigned i=0; i<program_state.instance->count_labels(); ++i)
				if(!program_state.solution->get_active(i).in(rot_angle))
					draw_label(cr, i, true, rot_angle);
		for(unsigned i=0; i<program_state.instance->count_labels(); ++i)
			if(program_state.solution->get_active(i).in(rot_angle))
				draw_label(cr, i, false, rot_angle);

	}

	cairo_restore(cr);

	if(draw_camera_mid){
		cairo_set_line_width (cr, 5.0);
		cairo_set_source_rgb (cr, 0, 0, 1);

		const int s = 7;
		cairo_move_to(cr, -s, -s);
		cairo_line_to(cr, +s, +s);
		cairo_move_to(cr, -s, +s);
		cairo_line_to(cr, +s, -s);
		cairo_stroke(cr);
	}
}

void take_screenshot()
{

	cairo_t *cr;
	cairo_surface_t *surface;

	surface =  cairo_image_surface_create(CAIRO_FORMAT_ARGB32, 800, 600);
	cr = cairo_create(surface);

	draw_all(cr, 800, 600);


	std::string filename;
	std::stringstream out;
	out << screenshot_counter++;
	filename = "image" + out.str() + ".png";

	cairo_surface_write_to_png(surface, filename.c_str());

	cairo_surface_destroy(surface);
	printf("screenshot %d taken\n", screenshot_counter);
	cairo_destroy(cr);
}


gint expose_event (GtkWidget *widget, GdkEventExpose *event)
{
	cairo_t *cr;
	cr = gdk_cairo_create(widget->window);

	gint w, h;
	gdk_drawable_get_size(gtk_widget_get_window(widget), &w, &h);

	draw_all(cr, w, h);

	cairo_destroy(cr);

	if (b_take_screenshots) take_screenshot();
	return FALSE;
}

void queue_draw(){
	if(frame_limiter){
		// Redraw at most 10 times per second
		static timeval last_time = {0, 0};
		timeval now;
		gettimeofday(&now, 0);

		if(last_time.tv_sec < now.tv_sec || last_time.tv_usec+100*1000 < now.tv_usec){
			last_time = now;
			gtk_widget_queue_draw(drawing_area);
		}
	}else
		gtk_widget_queue_draw(drawing_area);
}

void redraw_as_soon_as_possible(){
	gtk_widget_queue_draw(drawing_area);
}

// ------------------------------------------------------ //

double start_rotation_angle;

enum Action{
	ac_none,
	ac_rotate_map,
	ac_move_map,
	ac_move_label,
	ac_resize_label_left,
	ac_resize_label_right,
	ac_resize_label_bottom,
	ac_resize_label_top,
	ac_new_label
}action;

bool action_modifies_solution(Action a){
	return a == ac_move_label || a == ac_resize_label_left || a == ac_resize_label_right ||
		a == ac_resize_label_bottom || a == ac_resize_label_top || a == ac_new_label;
}

static void reset_state(){
	action = ac_none;
	selected_label = no_label;
}

bool is_no_solution_loaded(){
	if(program_state.solution){
		GtkWidget*dlg = gtk_message_dialog_new(GTK_WINDOW(window), GTK_DIALOG_DESTROY_WITH_PARENT, 
			GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
			"Can not edit this part of the instance while a solution is loaded. Unload it first.");
		gtk_window_set_title(GTK_WINDOW(dlg), "Error");
		gtk_dialog_run(GTK_DIALOG(dlg));
		gtk_widget_destroy(dlg);
		return false;
	}else
		return true;
}

bool cursor_over_label(unsigned i, int x, int y){
	const Label&l = program_state.instance->get_label(i);
	Vector v = world_trans*l.pos;
	return(v.x - l.left <= x && x <= v.x + l.right && v.y - l.top <= y && y <= v.y + l.bottom);
}

int map_move_start_x, map_move_start_y;

gint button_press_event (GtkWidget *widget, GdkEventButton *event)
{
	// Get cursor position & drawing area size
	int x, y;
	GdkModifierType state;
	gdk_window_get_pointer(event->window, &x, &y, &state);
	gint w, h;
	gdk_drawable_get_size(gtk_widget_get_window(widget), &w, &h);
	x -= w/2;
	y -= h/2;
	camera_to_map(x, y);

	if(event->button == 3) {
		map_move_start_x = x;
		map_move_start_y = y;
		action = ac_move_map;
	}else if(event->button == 1) {
		if(action == ac_new_label){
			Label l;
			l.pos = geo_util::inv(world_trans)*Vector(x,y);
			const int s = 20;
			l.left = 0;
			l.right = 2*s;
			l.bottom = 0;
			l.top = 2*s;
			l.text = "new";
			selected_label = program_state.instance->add_label(l);

			action = ac_none;
			redraw_as_soon_as_possible();
			return TRUE;
		}

		// Should the selected label be resized/moved?
		if(selected_label != no_label){
			const Label&l = program_state.instance->get_label(selected_label);
			Vector v = world_trans*l.pos;

			const int s = 5;

			if(x-s <= v.x && v.x <= x+s && y-s <= v.y && v.y <= y+s){
				action = ac_move_label;
			}

			if(x-s <= v.x - l.left && v.x - l.left <= x+s && 
				v.y - l.top < y && y < v.y + l.bottom){
				action = ac_resize_label_left;
			}

			if(x-s <= v.x + l.right && v.x + l.right <= x+s && 
				v.y - l.top < y && y < v.y + l.bottom){
				action = ac_resize_label_right;
			}

			if(y-s <= v.y - l.top && v.y - l.top <= y+s && 
				v.x - l.left < x && x < v.x + l.right){
				action = ac_resize_label_top;
			}

			if(y-s <= v.y + l.bottom && v.y + l.bottom <= y+s && 
				v.x - l.left < x && x < v.x + l.right){
				action = ac_resize_label_bottom;
			}

			if(action_modifies_solution(action)){
				if(!is_no_solution_loaded())
					action = ac_none;
				return TRUE;
			}
		}

		// Was a label selected ? 

		bool was_label_selected = false;

		if(selected_label == no_label){
			for(unsigned i=0; i<program_state.instance->count_labels(); ++i)
				if(cursor_over_label(i, x, y)){
					selected_label = i;
					was_label_selected = true;
				}
		}else{
			for(unsigned i=selected_label+1; i<program_state.instance->count_labels(); ++i)
				if(cursor_over_label(i, x, y)){
					selected_label = i;
					was_label_selected = true;
					break;
				}
			if(!was_label_selected){
				for(unsigned i=0; i<=selected_label; ++i)
					if(cursor_over_label(i, x, y)){
						selected_label = i;
						was_label_selected = true;
						break;
					}
				if(!was_label_selected){
					selected_label = no_label;
					redraw_as_soon_as_possible();
				}
			}
		}

		if(was_label_selected){
			redraw_as_soon_as_possible();
			return TRUE;
		}

		// If nothing more fancy was requested the map should be rotated
		start_rotation_angle = std::atan2(y, x);
		action = ac_rotate_map;

	}
	return TRUE;
}

static gint button_release_event (GtkWidget *widget, GdkEventButton *event)
{
	if(event->button == 1 || event->button == 3){
		action = ac_none;
		redraw_as_soon_as_possible();
	}
	return TRUE;
}


static gint motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
{
	if(action == ac_none || action == ac_new_label)
		return TRUE;
	
	int x, y;
	GdkModifierType unused;
	if (event->is_hint)
		gdk_window_get_pointer(event->window, &x, &y, &unused);
	else{
		x = event->x;
		y = event->y;
	}
	gint w, h;
	gdk_drawable_get_size(gtk_widget_get_window(widget), &w, &h);
	x -= w/2;
	y -= h/2;
	camera_to_map(x, y);

    if(action == ac_move_map){
        int dx = x-map_move_start_x, dy = y-map_move_start_y;
        map_move_start_x = x;
        map_move_start_y = y;
        world_trans = geo_util::translate(Vector(dx, dy))*world_trans;
    }




    if(action == ac_rotate_map){
        double temp = std::atan2(y,x);
        world_trans = geo_util::rot(temp-start_rotation_angle) * world_trans;
        start_rotation_angle = temp;
    }

    if(action == ac_move_label){
        Label&l = program_state.instance->get_label(selected_label);
        l.pos = geo_util::inv(world_trans)*Vector(x, y);
    }

    const double s = 0;

    if(action == ac_resize_label_left){
        Label&l = program_state.instance->get_label(selected_label);
        Vector v = world_trans*l.pos;
        l.left = std::max(s, v.x - x);
    }

    if(action == ac_resize_label_top){
        Label&l = program_state.instance->get_label(selected_label);
        Vector v = world_trans*l.pos;
        l.top = std::max(s, v.y - y);
    }

    if(action == ac_resize_label_right){
        Label&l = program_state.instance->get_label(selected_label);
        Vector v = world_trans*l.pos;
        l.right = std::max(s, x - v.x);
    }

    if(action == ac_resize_label_bottom){
        Label&l = program_state.instance->get_label(selected_label);
        Vector v = world_trans*l.pos;
        l.bottom = std::max(s, y - v.y);
    }

	queue_draw();

	return TRUE;
}

// -------------------------------------------------------- //

void error_dlg(const string&message){
	GtkWidget*dlg = gtk_message_dialog_new(GTK_WINDOW(window), GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
		"%s", message.c_str());
	gtk_window_set_title(GTK_WINDOW(dlg), "Error");
	gtk_dialog_run(GTK_DIALOG(dlg));
	gtk_widget_destroy(dlg);
}

void info_dlg(const string&message){
	GtkWidget*dlg = gtk_message_dialog_new(GTK_WINDOW(window), GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "%s", message.c_str());
	gtk_window_set_title(GTK_WINDOW(dlg), "Information");
	gtk_dialog_run(GTK_DIALOG(dlg));
	gtk_widget_destroy(dlg);
}

bool get_string_dlg(const string&message, string&text){
	GtkWidget*dlg = gtk_dialog_new();
	GtkWidget*label = gtk_label_new(message.c_str());
	GtkWidget*entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(entry), text.c_str());
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);

	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dlg)->vbox), label, FALSE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dlg)->vbox), entry, FALSE, TRUE, 5);

	gtk_dialog_add_button(GTK_DIALOG(dlg), GTK_STOCK_OK, GTK_RESPONSE_OK);
	gtk_dialog_add_button(GTK_DIALOG(dlg), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	
	gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_OK);

	gtk_widget_show_all(dlg);

	if(gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_OK){
		text = gtk_entry_get_text(GTK_ENTRY(entry));
		gtk_widget_destroy(dlg);
		return true;
	}else{
		gtk_widget_destroy(dlg);
		return false;
	}
}

string select_a_label_message = "A label must be selected to perform this action.";

// -------------------------------------------------------- //

const int rotation_360_steps = 400;
int rotation_360_count_down = 0;

gboolean map_rotation_360_timer_callback(gpointer){
	world_trans = geo_util::rot(2*M_PI/rotation_360_steps)*world_trans;
	redraw_as_soon_as_possible();
	return rotation_360_count_down-- != 0;
}

void on_start_map_rotation_360(){
	rotation_360_count_down = rotation_360_steps;
	g_timeout_add(33, map_rotation_360_timer_callback, 0);
}

gboolean label_rotation_360_timer_callback(gpointer){
	world_trans = geo_util::rot(2*M_PI/rotation_360_steps)*world_trans;
	camera_trans = geo_util::rot(-2*M_PI/rotation_360_steps)*camera_trans;
	redraw_as_soon_as_possible();
	return rotation_360_count_down-- != 0;
}

void on_start_label_rotation_360(){
	rotation_360_count_down = rotation_360_steps;
	g_timeout_add(33, label_rotation_360_timer_callback, 0);
}

gboolean all_rotation_360_timer_callback(gpointer){
	camera_trans = geo_util::rot(2*M_PI/rotation_360_steps)*camera_trans;
	redraw_as_soon_as_possible();
	return rotation_360_count_down-- != 0;
}

void on_start_all_rotation_360(){
	rotation_360_count_down = rotation_360_steps;
	g_timeout_add(33, all_rotation_360_timer_callback, 0);
}

// -------------------------------------------------------- //

struct FlightPoint{
	FlightPoint(){}
	FlightPoint(Vector target, double angle_speed, double move_speed):
		target(target), angle_speed(angle_speed), move_speed(move_speed){}

	Vector target;
	double angle_speed;
	double move_speed;
};

vector<FlightPoint>flight_point_queue;

bool is_nearly_null(double x){
	const double eps = 0.0001;
	return -eps <= x && x <= eps;
}

gboolean flight_timer_callback(gpointer){
	if(flight_point_queue.empty())
		return FALSE;


	const double angle_speed = flight_point_queue.back().angle_speed*M_PI/100.0;
	const double move_speed = flight_point_queue.back().move_speed*3;
	Vector t = world_trans*flight_point_queue.back().target;

	if(is_nearly_null(t.x)){
		if(is_nearly_null(t.y)){
			flight_point_queue.pop_back();
			return TRUE;
		}else{
			if(-t.y < move_speed){
				world_trans = geo_util::translate(t)*world_trans;
				t.y = 0.0;
			}else{
				t.y += move_speed;
				world_trans = geo_util::translate(Vector(0, move_speed))*world_trans;
			}
		}
	}else{
		double angle = normalize_angle(atan2(t.y, t.x)-M_PI/2)-M_PI;
		if(angle < -angle_speed){
			t = geo_util::rot(+angle_speed)*t;
			world_trans = geo_util::rot(+angle_speed)*world_trans;
		}else if(angle > angle_speed){
			t = geo_util::rot(-angle_speed)*t;
			world_trans = geo_util::rot(-angle_speed)*world_trans;
		}else{
			t = Vector(0.0, -norm(t));
			world_trans = geo_util::rot(-angle)*world_trans;
		}
	}

	flight_point_queue.back().target = geo_util::inv(world_trans)*t;
	redraw_as_soon_as_possible();
	return TRUE;
}

bool load_flight(const char*filename){
	FILE*f = fopen(filename, "r");
	if(!f)
		return false;

	char buffer[1024];
	fgets(buffer, sizeof(buffer)-1, f); // ignore first line that contains column names

	vector<FlightPoint>temp;
	double x, y, angle_speed, move_speed;
	while(fscanf(f, "%lf%lf%lf%lf", &x, &y, &angle_speed, &move_speed) == 4)
		temp.push_back(FlightPoint(Vector(x, y), angle_speed, move_speed));
	fclose(f);
	if(temp.empty())
		return false;
	temp.swap(flight_point_queue);
	reverse(flight_point_queue.begin(), flight_point_queue.end());
	g_timeout_add(33, flight_timer_callback, 0);
	return true;
}

void load_flight_dlg(){
	GtkWidget *dlg = gtk_file_chooser_dialog_new ("Load flight",
				GTK_WINDOW(window),
				GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				NULL);
	if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dlg));
		if(!load_flight(filename))
			error_dlg("Could not load flight");
		g_free (filename);
	}
	gtk_widget_destroy (dlg);
}

// -------------------------------------------------------- //

void sync_background(){
	if(program_state.background_filename.empty()){
		if(background)
			cairo_surface_destroy(background);
		background = 0;
	}else{
		cairo_surface_t*new_background = cairo_image_surface_create_from_png(program_state.background_filename.c_str());

		string error_message;
		switch(cairo_surface_status(new_background)){
		case CAIRO_STATUS_FILE_NOT_FOUND:
			error_message = "Could find png file";
			break;
		case CAIRO_STATUS_READ_ERROR:
			error_message = "Error reading png file";
			break;
		case CAIRO_STATUS_NO_MEMORY:
			error_message = "Not enough memory to read png file";
			break;

		}

		if(!error_message.empty()){
			cairo_surface_destroy(new_background);
			error_message += ' ';
			error_message += program_state.background_filename;
			error_dlg(error_message);
		}else{
			if(background)
				cairo_surface_destroy(background);
			background = new_background;
		}
	}
	if(drawing_area)
		redraw_as_soon_as_possible();
}


void sync_font(){
	if(program_state.font_desc.empty()){
		if(font_desc)
			pango_font_description_free(font_desc);
		font_desc = 0;
	}else{
		PangoFontDescription *temp = pango_font_description_from_string(program_state.font_desc.c_str());
		if(!temp){
			error_dlg("Could not load font "+program_state.font_desc);
			return;
		}
		if(font_desc) {
			pango_font_description_free(font_desc);
		}
		font_desc = temp;
	}
	if(drawing_area)
		redraw_as_soon_as_possible();
}

void sync_resource_cache(){
	sync_background();
	sync_font();
}

// -------------------------------------------------------- //

void select_font_dlg(){
	GtkWidget * dlg = gtk_font_selection_dialog_new("Select Label Font");
	if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_OK)
	{
		char *temp = gtk_font_selection_dialog_get_font_name (GTK_FONT_SELECTION_DIALOG (dlg));
		program_state.font_desc = temp;
		g_free (temp);
		sync_font();
	}
	gtk_widget_destroy (dlg);
}
// -------------------------------------------------------- //


void load_program_state(const string&filename){
	if(!load(filename, program_state))
		error_dlg("Can not open "+filename);
	else
		sync_resource_cache();
}

static void load_dlg(){
	GtkWidget *dlg = gtk_file_chooser_dialog_new ("Open File",
				      GTK_WINDOW(window),
				      GTK_FILE_CHOOSER_ACTION_OPEN,
				      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				      GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				      NULL);
	if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dlg));
		load_program_state(filename);
		save_to_file = filename;
		g_free (filename);
		sync_font();
	}
	gtk_widget_destroy (dlg);
}

static void save_as_dlg(){
	GtkWidget *dlg = gtk_file_chooser_dialog_new ("Save File",
					      GTK_WINDOW(window),
					      GTK_FILE_CHOOSER_ACTION_SAVE,
					      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					      GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					      NULL);
	gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dlg), TRUE);

	/*if (user_edited_a_new_document)
	  {
	    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), default_folder_for_saving);
	    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "Untitled document");
	  }
	else
	  gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), filename_for_existing_document);*/

	if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename;
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dlg));
		save(filename, program_state);
		save_to_file = filename;
		g_free (filename);
	}
	gtk_widget_destroy (dlg);

}

static void save_dlg(){
	if(save_to_file.empty())
		save_as_dlg();
	else
		save(save_to_file, program_state);
		
}

// ------------------------------------------------------ //

// http://library.gnome.org/devel/gtk/unstable/GtkFileChooserDialog.html
// http://www.gtk.org/api/2.6/gtk/GtkMessageDialog.html#gtk-message-dialog-new

static void on_new_instance()
{
	reset_state();
	program_state.background_filename = "";
	program_state.font_desc = "";
	delete program_state.solution;
	program_state.solution = 0;
	delete program_state.instance;
	program_state.instance = new Instance;
	sync_resource_cache();
	save_to_file = "";
	world_trans = geo_util::id();
	camera_trans = geo_util::id();
}

void select_background_dlg(){
	GtkWidget *dlg = gtk_file_chooser_dialog_new ("Select background",
				      GTK_WINDOW(window),
				      GTK_FILE_CHOOSER_ACTION_OPEN,
				      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				      GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				      NULL);

	GtkFileFilter *filter =  gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "PNG");
	gtk_file_filter_add_mime_type(filter, "image/png");
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dlg), filter);

	if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dlg));

		program_state.background_filename = filename;
		sync_background();
		g_free (filename);


	}
	gtk_widget_destroy (dlg);
}

static void on_place_new_label()
{
	if(is_no_solution_loaded()){
		action = ac_new_label;
		selected_label = no_label;
		redraw_as_soon_as_possible();
	}
}

static void on_remove_label()
{
	if(is_no_solution_loaded()){
		if(selected_label == no_label){
			info_dlg(select_a_label_message);
		}else{
			program_state.instance->remove_label(selected_label);
			selected_label = no_label;
			redraw_as_soon_as_possible();
		}
	}
}

// ------------------------------------------------------ //


static void rename_label_dlg()
{
	if(selected_label == no_label){
		info_dlg(select_a_label_message);
	}else{
		if(get_string_dlg("Enter the name of the label:", program_state.instance->get_label(selected_label).text))
			redraw_as_soon_as_possible();
	}
}

// ------------------------------------------------------ //

const double key_rotation_angle = M_PI/90;

void on_world_left(){
	world_trans = geo_util::rot(-key_rotation_angle)*world_trans;
	redraw_as_soon_as_possible();
}

void on_world_right(){
	world_trans = geo_util::rot(key_rotation_angle)*world_trans;
	redraw_as_soon_as_possible();
}

void on_camera_left(){
	camera_trans = geo_util::rot(-key_rotation_angle)*camera_trans;
	redraw_as_soon_as_possible();
}

void on_camera_right(){
	camera_trans = geo_util::rot(key_rotation_angle)*camera_trans;
	redraw_as_soon_as_possible();
}

void on_camera_reset(){
	world_trans = geo_util::id();
	camera_trans = geo_util::id();
	redraw_as_soon_as_possible();
}

void on_print_active_time(){
  if (program_state.solution == NULL)
    return;

  cout << endl << "total active time:  " << program_state.solution->active_time() << endl;
}

void on_random_instance(){
  if (program_state.solution != NULL){
	delete program_state.solution;
	program_state.solution = 0;
  }

  int nmb_labels = 60;
  int width = 1000;
  int height = 1000;

  //srand(12345);

  for (int i = 0; i < nmb_labels; i++){
    double x = rand()%width;
    double y = rand()%height;
    Label l;
    l.pos = Vector(x,y);
    const int s = 40;
    l.left = 0;
    l.right = s;
    l.bottom = 0;
    l.top = s;
    l.text = std::to_string(i);
    selected_label = program_state.instance->add_label(l);
    // std::cout << x << ", " << y << "   ";
  }
  //std::cout << std::endl;

  redraw_as_soon_as_possible();
}

void on_approx_scale(){
  Instance& inst =  *program_state.instance;
  double min_lat(std::numeric_limits<double>::max());
  double max_lat(std::numeric_limits<double>::min());
  double min_lon(std::numeric_limits<double>::max());
  double max_lon(std::numeric_limits<double>::min());

  double min_y(std::numeric_limits<double>::max());
  double max_y(std::numeric_limits<double>::min());
  double min_x(std::numeric_limits<double>::max());
  double max_x(std::numeric_limits<double>::min());

  for (unsigned int i = 0; i < inst.count_labels(); i++) {
    Label& l = inst.get_label(i);
    double x = l.pos.x;
    double y = l.pos.y;

    min_x = std::min(x, min_x);
    max_x = std::max(x, max_x);
    min_y = std::min(y, min_y);
    max_y = std::max(y, max_y);
  }

  double delta_x_px = max_x - min_x;
  double delta_y_px = max_y - min_y;

  for (unsigned int i = 0; i < inst.count_labels(); i++) {
    Label& l = inst.get_label(i);
    double x = l.pos.x;
    double y = l.pos.y;
    double lat = y2lat(y * delta_y_px);
    double lon = x2lon(x * delta_x_px);

    min_lat = std::min(min_lat, lat);
    max_lat = std::max(max_lat, lat);
    min_lon = std::min(min_lon, lon);
    max_lon = std::max(max_lon, lon);
  }

  cout << min_lat << ", " << max_lat << ", " << min_lon << ", " << max_lon << endl;

  double delta_x_m1 = latlongdist(max_lat, min_lon, max_lat, max_lon);
  double delta_x_m2 = latlongdist(min_lat, min_lon, min_lat, max_lon);

  double delta_y_m1 = latlongdist(min_lat, min_lon, max_lat, min_lon);
  double delta_y_m2 = latlongdist(min_lat, max_lon, max_lat, max_lon);

  cout << "distances:" << endl;
  cout << delta_x_m1 << " " << delta_x_m2 << endl;
  cout << delta_y_m1 << " " << delta_y_m2 << endl;


}

void on_label_left(){
	world_trans = geo_util::rot(key_rotation_angle)*world_trans;	
	on_camera_left();
}
void on_label_right(){
	world_trans = geo_util::rot(-key_rotation_angle)*world_trans;	
	on_camera_right();
}

void on_flip_take_screenshots()
{
	printf("take screenshots: %s", ((b_take_screenshots)? "true": "false") );
	b_take_screenshots = !b_take_screenshots;
	if (not b_take_screenshots)
		screenshot_counter = 0;
}

void on_flip_draw_rectangles()
{
	draw_label_rectangle = not  draw_label_rectangle;
	redraw_as_soon_as_possible();
}

void on_get_rot_angle()
{
	double rot_angle = geo_util::rot_angle(world_trans.m);
	double rot = (rot_angle * 180.0 / 3.141);
	cout <<  (rot < 0 ? 360.0 + rot : rot)   << endl;
}

void on_move_to()
{
  double dx = 0;
  double dy = 0;
  if ( program_state.instance->count_labels() > 0){
    dx = program_state.instance->get_label(0).pos.x;
    dy = program_state.instance->get_label(0).pos.y;
  }
  world_trans = geo_util::id();

  world_trans = geo_util::translate(Vector(-dx, -dy));
        
  redraw_as_soon_as_possible();
}

void on_load_flight_dlg(){
	load_flight("test.txt");
}


template<AlgorithmInfo::AlgorithmName name, int nmb_intervals, bool use_hard_conflicts, bool presolve>
void solve_mip_fast() {
    if (!program_state.instance){
        error_dlg("No instance loaded");
    } else {
        delete program_state.solution;

        int new_nmb_intervals = nmb_intervals;
        if (new_nmb_intervals < 0) {
            string max_intervals = "2";
            if(get_string_dlg("max #intervals ", max_intervals)){
              new_nmb_intervals = atoi(max_intervals.c_str());

              if (new_nmb_intervals == 0)
                new_nmb_intervals = -1;
            }
        }

        AlgorithmInfo algo;
        algo.name = name;
        algo.max_nmb_intervals = new_nmb_intervals;
        algo.use_hard_conflicts = use_hard_conflicts;

        program_state.solution = new Solution(calc_mip_solution_fast(algo, *program_state.instance, presolve));

        reset_state();
        redraw_as_soon_as_possible();
    }

}

void on_solve_mip_no_flickering() {

	delete program_state.solution;
	program_state.solution = new Solution(calc_mip_solution_no_flickering(*program_state.instance));
	reset_state();
	redraw_as_soon_as_possible();
}


template<AlgorithmInfo::AlgorithmName name, int nmb_intervals, bool use_hard_conflicts>
void on_calc_greedy_solution()
{
    if (!program_state.instance){
        error_dlg("No instance loaded");
    } else {
        delete program_state.solution;

        int new_nmb_intervals = nmb_intervals;
        if (new_nmb_intervals < 0) {
            string max_intervals = "2";
            if(get_string_dlg("max #intervals ", max_intervals)){
              new_nmb_intervals = atoi(max_intervals.c_str());

              if (new_nmb_intervals == 0)
                new_nmb_intervals = -1;
            }
        }

        AlgorithmInfo algo;
        algo.name = name;
        algo.max_nmb_intervals = new_nmb_intervals;
        algo.use_hard_conflicts = use_hard_conflicts;

        program_state.solution = new Solution(calc_greedy(*program_state.instance, algo));

        reset_state();
        redraw_as_soon_as_possible();
    }
}



// ------------------------------------------------------ //

//

template<bool quarter_apx, int max_nmb_intervals, bool use_hard_conflicts, AlgorithmInfo::AlgorithmName greedy_alg = AlgorithmInfo::UNSPECIFIED>
void solve_appr() {
    if(!program_state.instance){
        error_dlg("No instance loaded");
    }else{
        int app = 2;
        if (!quarter_apx) {
            string s = "2";
            if(get_string_dlg("Enter ", s)){
                int i = atoi(s.c_str());
                if(i < 2) {
                    error_dlg(s+" must be at least 2");
                    app = 2;
                }
                else{
                    app = i;
                }
            }
        }

        int nmb_intervals = max_nmb_intervals;
        if (nmb_intervals != 1) {
            string max_intervals = "2";
            if(get_string_dlg("max #intervals ", max_intervals)){
              nmb_intervals = atoi(max_intervals.c_str());
              if (nmb_intervals == 0)
                nmb_intervals = -1;
            }
        }

        delete program_state.solution;
        program_state.solution = new Solution(calc_app(*program_state.instance, nmb_intervals, use_hard_conflicts, app * app, greedy_alg));
        reset_state();
        redraw_as_soon_as_possible();
    }
}

void move_background(){
    string s = "0";
    if(get_string_dlg("move by x", s)){
      int dx = atoi(s.c_str());

      s = "0";
      if(get_string_dlg("move by y", s)){
	int dy = atoi(s.c_str());
	
	background_x += dx;
	background_y += dy;
	
	std::cout << "background move: " << background_x << ", " << background_y << std::endl;

	redraw_as_soon_as_possible();
      }
    }
}


// ---- //

void on_snap_to_left(){
	if(selected_label == no_label)
		info_dlg(select_a_label_message);
	else{
		Label&l = program_state.instance->get_label(selected_label);
		l.right += l.left;
		l.left = 0;
		redraw_as_soon_as_possible();
	}
}

void on_snap_to_right(){
	if(selected_label == no_label)
		info_dlg(select_a_label_message);
	else{
		Label&l = program_state.instance->get_label(selected_label);
		l.left += l.right;
		l.right = 0;
		redraw_as_soon_as_possible();
	}
}

void on_snap_to_top(){
	if(selected_label == no_label)
		info_dlg(select_a_label_message);
	else{
		Label&l = program_state.instance->get_label(selected_label);
		l.bottom += l.top;
		l.top = 0;
		redraw_as_soon_as_possible();
	}
}

void on_snap_to_bottom(){
	if(selected_label == no_label)
		info_dlg(select_a_label_message);
	else{
		Label&l = program_state.instance->get_label(selected_label);
		l.top += l.bottom;
		l.bottom = 0;
		redraw_as_soon_as_possible();
	}
}

// ------------------------------------------------------ //

void resize_label_to_text(int index){
    if(label_text_width.size() <= static_cast<unsigned int>(index) || label_text_height.size() <= static_cast<unsigned int>(index))
		return;

	const int s = 4;

	Label&l = program_state.instance->get_label(index);
	{
		int old_height = l.top + l.bottom;
		int new_height = label_text_height[index] + 2*s;
		l.top *= new_height;
		l.top /= old_height;
		l.bottom *= new_height;
		l.bottom /= old_height;
	}
}

void on_resize_label_to_text(){
	if(selected_label != no_label){
		resize_label_to_text(selected_label);
		redraw_as_soon_as_possible();
	}else
		info_dlg(select_a_label_message);
}

void on_resize_all_labels_to_text(){
    for(unsigned int i=0; i<program_state.instance->count_labels(); ++i)
		resize_label_to_text(i);
	redraw_as_soon_as_possible();
}

// ------------------------------------------------------ //

static void on_hide_or_transparent(){
	draw_hidden_labels_transparent = !draw_hidden_labels_transparent;
	redraw_as_soon_as_possible();
}

static void on_hide_solution_circle(){
	hide_solution_circle = !hide_solution_circle;
	redraw_as_soon_as_possible();
}

static void on_draw_camera_mid(){
	draw_camera_mid = !draw_camera_mid;
	redraw_as_soon_as_possible();
}

static void on_frame_limiter(){
	frame_limiter = !frame_limiter;
}

static void on_clear_solution()
{
	delete program_state.solution;
	program_state.solution = 0;
	redraw_as_soon_as_possible();
}

template<Solution (*calc_solution)(const Instance&)>
void on_calc_solution()
{
	if(!program_state.instance){
		error_dlg("No instance loaded");
	}else{
		delete program_state.solution;
		program_state.solution = new Solution(calc_solution(*program_state.instance));
		
		
		reset_state();
		redraw_as_soon_as_possible();
	}
}

// ------------------------------------------------------ //

static gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return false;
}

static void destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

// ------------------------------------------------------ //

void build_window(){
	// Make window
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Viewer");
	gtk_widget_set_usize (GTK_WIDGET(window), 800, 600);
	g_signal_connect(window, "delete-event", G_CALLBACK(delete_event), NULL);
	g_signal_connect(window, "destroy", G_CALLBACK(destroy), NULL);

	// Make drawing area
	drawing_area = gtk_drawing_area_new();
	//gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 800, 600);

	gtk_widget_set_events (drawing_area, 
			GDK_EXPOSURE_MASK|GDK_LEAVE_NOTIFY_MASK|GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|GDK_POINTER_MOTION_MASK|GDK_POINTER_MOTION_HINT_MASK);

	g_signal_connect(drawing_area, "expose-event", G_CALLBACK (expose_event), NULL);
	g_signal_connect(drawing_area, "button-press-event", G_CALLBACK (button_press_event), NULL);
	g_signal_connect(drawing_area, "button-release-event", G_CALLBACK (button_release_event), NULL);
	g_signal_connect(drawing_area, "motion-notify-event", G_CALLBACK (motion_notify_event), NULL);
	gtk_widget_show(drawing_area);

	// http://www.gtk.org/tutorial1.2/gtk_tut-13.html

	// Make menu bar

	static GtkItemFactoryEntry menu_items[] = {
		{ "/_File",                  NULL,         NULL                    , 0, "<Branch>" },
		{ "/File/_New",              "<control>N",         on_new_instance         , 0, NULL },
		{ "/File/_Open",             "<control>O", load_dlg                , 0, NULL },
		{ "/File/_Save",             "<control>S", save_dlg                , 0, NULL },
		{ "/File/Save _As",          NULL,         save_as_dlg             , 0, NULL },
		{ "/File/sep1",              NULL,         NULL                    , 0, "<Separator>" },
		{ "/File/Set _Background",   NULL,         select_background_dlg   , 0, NULL },
		{ "/File/Start Flight",      NULL,         load_flight_dlg         , 0, NULL },
		{ "/File/sep2",              NULL,         NULL                    , 0, "<Separator>" },
		{ "/File/Do not limit frame rate",
			NULL,         on_frame_limiter        , 0, "<CheckItem>" },
		{ "/File/Draw position",     NULL,         on_draw_camera_mid      , 0, "<CheckItem>" },
		{ "/File/sep3",              NULL,         NULL                    , 0, "<Separator>" },
		{ "/File/Quit",              "<alt>F4",    gtk_main_quit           , 0, NULL },
		{ "/_Label",                 NULL,         NULL                    , 0, "<Branch>" },
		{ "/Label/New",              "<control>A", on_place_new_label      , 0, NULL },
		{ "/Label/Delete",           NULL,         on_remove_label         , 0, NULL },
		{ "/Label/sep3",             NULL,         NULL                    , 0, "<Separator>" },
		{ "/Label/Rename",           "<control>M", rename_label_dlg        , 0, NULL },
		{ "/Label/Select font",      NULL,         select_font_dlg         , 0, NULL },
		{ "/Label/Resize label to text",
			"<control>K", on_resize_label_to_text , 0, NULL },
		{ "/Label/Resize all labels to text",
			NULL, on_resize_all_labels_to_text    , 0, NULL },
		{ "/Label/sep3",             NULL,         NULL                    , 0, "<Separator>" },
		{ "/Label/Snap to left",     "<control>L", on_snap_to_left         , 0, NULL },
		{ "/Label/Snap to top",      "<control>T", on_snap_to_top          , 0, NULL },
		{ "/Label/Snap to right",    "<control>R", on_snap_to_right        , 0, NULL },
		{ "/Label/Snap to bottom",   "<control>B", on_snap_to_bottom       , 0, NULL },
		{ "/_Rotate",                NULL, NULL                            , 0, "<Branch>" },
		{ "/Rotate/World left",      "a", on_world_left                    , 0, NULL },
		{ "/Rotate/World right",     "d", on_world_right                   , 0, NULL },
		{ "/Rotate/Camera left",     "q", on_camera_left                   , 0, NULL },
		{ "/Rotate/Camera right",    "e", on_camera_right                  , 0, NULL },
		{ "/Rotate/Reset",           NULL, on_camera_reset                 , 0, NULL },
		{ "/Rotate/Rotate map 360",  "1", on_start_map_rotation_360       , 0, NULL },
		{ "/Rotate/Rotate label 360","2", on_start_label_rotation_360     , 0, NULL },
		{ "/Rotate/Rotate all 360",  "3",   on_start_all_rotation_360     , 0, NULL },
		{ "/Rotate/Label left", 	"x", on_label_left, 0, NULL},
		{ "/Rotate/Label right", 	"y", on_label_right, 0, NULL},
		{ "/_Solution",              NULL,         NULL                    , 0, "<Branch>" },
		{ "/Solution/Hide labels (instead of making them transparent)",      
			NULL,         on_hide_or_transparent  , 0, "<CheckItem>" },
		{ "/Solution/Hide Solution Circles",      
			NULL,         on_hide_solution_circle , 0, "<CheckItem>" },
		{ "/Solution/Draw Rectangles", "5",  on_flip_draw_rectangles   , 0, "<CheckItem>"},
		{ "/Solution/Clear solution",NULL,         on_clear_solution       , 0, NULL },
		{ "/Solution/sep3",          NULL,         NULL                    , 0, "<Separator>" },

		{ "/Solution/Solve by MIP",                  NULL,         NULL                    , 0, "<Branch>" },
        { "/Solution/Solve by MIP/MIP", "", solve_mip_fast<AlgorithmInfo::MIP, 1, false, false>, 0, NULL},
        { "/Solution/Solve by MIP/MIP presolve", "", solve_mip_fast<AlgorithmInfo::MIP, 1, false, true>, 0, NULL},
        { "/Solution/Solve by MIP/MIP inconsistent", "", solve_mip_fast<AlgorithmInfo::MIP, -1, false, false>, 0, NULL},
        { "/Solution/Solve by MIP/MIP w hard conflicts", "u", solve_mip_fast<AlgorithmInfo::MIP, 1, true, false>, 0, NULL},
        { "/Solution/Solve by MIP/MIP inconsistent w hard conflicts", "", solve_mip_fast<AlgorithmInfo::MIP, -1, true, false>, 0, NULL},
		{ "/Solution/Solve by MIP/sep1", NULL, NULL, 0, "<Separator>"},
		{ "/Solution/Solve by MIP/no Flickering", NULL, on_solve_mip_no_flickering, 0, NULL},


		{ "/Solution/Solve by MIP/sep2", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by MIP/MIP (alternative)", "", solve_mip_fast<AlgorithmInfo::MIP2, 1, false, false>, 0, NULL},
        { "/Solution/Solve by MIP/MIP (alternative) presolve", "", solve_mip_fast<AlgorithmInfo::MIP2, 1, false, true>, 0, NULL},
        { "/Solution/Solve by MIP/MIP (alternative) w hard conflicts", "", solve_mip_fast<AlgorithmInfo::MIP2, 1, true, false>, 0, NULL},

		{ "/Solution/Solve by approximation",                  NULL,         NULL                    , 0, "<Branch>" },
        { "/Solution/Solve by approximation/quarter approximation", NULL, solve_appr<true, 1, false>, 0, NULL},
        { "/Solution/Solve by approximation/quarter approximation w hard conflicts", NULL, solve_appr<true, 1, true>, 0, NULL},

        { "/Solution/Solve by approximation/apx", NULL, solve_appr<true, 1, false>, 0, NULL},
		{ "/Solution/Solve by approximation/sep1", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by approximation/plus greedy fill max", NULL, solve_appr<false, 1, false, AlgorithmInfo::GreedyMax>, 1, NULL},
        { "/Solution/Solve by approximation/plus greedy fill low cost", NULL, solve_appr<false, 1, false, AlgorithmInfo::GreedyLowCost>, 2, NULL},
        { "/Solution/Solve by approximation/plus greedy fill best ratio", NULL, solve_appr<false, 1, false, AlgorithmInfo::GreedyBestRatio>, 3, NULL},
		{ "/Solution/Solve by approximation/sep2", NULL, NULL, 0, "<Separator>"},

        { "/Solution/Solve by approximation/plus greedy fill max (inconsistent)", NULL, solve_appr<false, -1, false, AlgorithmInfo::GreedyMax>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill low cost (inconsistent)", NULL, solve_appr<false, -1, false, AlgorithmInfo::GreedyLowCost>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill best ratio (inconsistent)", NULL, solve_appr<false, -1, false, AlgorithmInfo::GreedyBestRatio>, 0, NULL},
		{ "/Solution/Solve by approximation/sep1", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by approximation/plus greedy fill max w hard conflicts", NULL, solve_appr<false, 1, true, AlgorithmInfo::GreedyMax>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill low cost w hard conflicts", NULL, solve_appr<false, 1, true, AlgorithmInfo::GreedyLowCost>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill best ratio w hard conflicts", NULL, solve_appr<false, 1, true, AlgorithmInfo::GreedyBestRatio>, 0, NULL},
		{ "/Solution/Solve by approximation/sep2", NULL, NULL, 0, "<Separator>"},

        { "/Solution/Solve by approximation/plus greedy fill max (inconsistent) w hard conflicts", NULL, solve_appr<false, -1, true, AlgorithmInfo::GreedyMax>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill low cost (inconsistent) w hard conflicts", NULL, solve_appr<false, -1, true, AlgorithmInfo::GreedyLowCost>, 0, NULL},
        { "/Solution/Solve by approximation/plus greedy fill best ratio (inconsistent) w hard conflicts", NULL, solve_appr<false, -1, true, AlgorithmInfo::GreedyBestRatio>, 0, NULL},

		{ "/Solution/Solve by Greedy", NULL, NULL, 0, "<Branch>" },
        { "/Solution/Solve by Greedy/Max", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyMax, 1, false>, 0, NULL},
        { "/Solution/Solve by Greedy/Low Cost", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyLowCost, 1, false>, 0, NULL},
        { "/Solution/Solve by Greedy/Best Ratio", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyBestRatio, 1, false>, 0, NULL},
		{ "/Solution/Solve by Greedy/sep1", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by Greedy/Max (inconsistent)", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyMax, -1, false>, 0, NULL},
        { "/Solution/Solve by Greedy/Low Cost (inconsistent)", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyLowCost, -1, false>, 0, NULL},
        { "/Solution/Solve by Greedy/Best Ratio(inconsistent)", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyBestRatio, -1, false>, 0, NULL},
		{ "/Solution/Solve by Greedy/sep2", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by Greedy/Max w hard conflicts", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyMax, 1, true>, 0, NULL},
        { "/Solution/Solve by Greedy/Low Cost  w hard conflicts", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyLowCost, 1, true>, 0, NULL},
        { "/Solution/Solve by Greedy/Best Ratio w hard conflicts", NULL, on_calc_greedy_solution<AlgorithmInfo::GreedyBestRatio, 1, true>, 0, NULL},
		{ "/Solution/Solve by Greedy/sep3", NULL, NULL, 0, "<Separator>"},
        { "/Solution/Solve by Greedy/Max (inconsistent) w hard conflicts", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyMax, -1, true>, 0, NULL},
        { "/Solution/Solve by Greedy/Low Cost (inconsistent) w hard conflicts", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyLowCost, -1, true>, 0, NULL},
        { "/Solution/Solve by Greedy/Best Ratio(inconsistent) w hard conflicts", NULL,  on_calc_greedy_solution<AlgorithmInfo::GreedyBestRatio, -1, true>, 0, NULL},

		{ "/Debug", NULL, NULL, 0, "<Branch>"},
		{ "/Debug/Take Screenshots", "4",  on_flip_take_screenshots   , 0, "<CheckItem>"},
		{ "/Debug/getRotangle", "g", on_get_rot_angle, 0, NULL},
		{ "/Debug/moveTo", "m", on_move_to, 0, NULL},
        { "/Debug/Screenshot", "s", take_screenshot, 0, NULL},
		{ "/Debug/Start Flight",    "6",         on_load_flight_dlg         , 0, NULL },
		{ "/Debug/Print Active Time",    "p",         on_print_active_time         , 0, NULL },
		{ "/Debug/Random Instance",    "r",         on_random_instance         , 0, NULL },
		{ "/Debug/approx scale",    "",         on_approx_scale         , 0, NULL },
		{ "/Debug/move background",    "g",         move_background        , 0, NULL },
		{ "/_Help",                  NULL,         NULL                    , 0, "<LastBranch>" },

		{ "/_Help/About",            NULL,         NULL                    , 0, NULL },
		{ "/_Help/Todo",             NULL,         NULL                    , 0, NULL }

	};

	GtkAccelGroup *accel_group = gtk_accel_group_new ();
	GtkItemFactory* item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", accel_group);
	gtk_item_factory_create_items (item_factory, sizeof (menu_items) / sizeof (menu_items[0]), menu_items, NULL);
	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
	GtkWidget *menubar = gtk_item_factory_get_widget (item_factory, "<main>");
	gtk_widget_show(menubar);

	// Put them together
	GtkWidget * main_vbox = gtk_vbox_new (FALSE, 1);
	gtk_container_border_width (GTK_CONTAINER (main_vbox), 1);
	gtk_container_add (GTK_CONTAINER (window), main_vbox);
	gtk_widget_show (main_vbox);

	gtk_box_pack_start (GTK_BOX(main_vbox), menubar, FALSE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(main_vbox), drawing_area, TRUE, TRUE, 0);

	frame_limiter = false;

	// Show the window
	gtk_widget_show(window);
}

// ------------------------------------------------------ //

void cleanup_globals(){
	if(background)
		cairo_surface_destroy(background);
	if(font_desc)
		pango_font_description_free(font_desc);
	delete program_state.solution;
	delete program_state.instance;
}

int main(int argc, char *argv[])
{
	putenv("LC_ALL=C");
	gtk_init (&argc, &argv);

	//setlocale(LC_NUMERIC, "C");


	on_new_instance();
	build_window();
	//handle_commandline_arguments(argc, argv);

	save_to_file = "";
	gtk_main ();

	cleanup_globals();
	srand(12345);

	redraw_as_soon_as_possible();


	return 0;
}

